from __future__ import annotations

import math
from pathlib import Path
import random
from typing import Protocol

import pyray as rl

from grim.audio import AudioState, update_audio
from grim.assets import PaqTextureCache, load_paq_entries
from grim.config import CrimsonConfig
from grim.fonts.grim_mono import GrimMonoFont, draw_grim_mono_text, load_grim_mono_font
from grim.fonts.small import SmallFontData, draw_small_text, load_small_font, measure_small_text_width

from grim.rand import Crand
from .game_world import GameWorld
from .gameplay import PlayerInput, PlayerState, weapon_assign_player
from .ui.cursor import draw_menu_cursor
from .weapons import WEAPON_TABLE, WeaponId, projectile_type_id_from_weapon_id, weapon_entry_for_projectile_type_id

WORLD_SIZE = 1024.0
DEMO_VARIANT_COUNT = 6

_DEMO_UPSELL_MESSAGES: tuple[str, ...] = (
    "Want more Levels?",
    "Want more Weapons?",
    "Want more Perks?",
    "Want unlimited Play time?",
    "Want to post your high scores?",
)

DEMO_PURCHASE_URL = "http://buy.crimsonland.com"
DEMO_PURCHASE_SCREEN_LIMIT_MS = 16_000
DEMO_PURCHASE_INTERSTITIAL_LIMIT_MS = 10_000

_DEMO_PURCHASE_TITLE = "Upgrade to the full version of Crimsonland Today!"
_DEMO_PURCHASE_FEATURES_TITLE = "Full version features:"
_DEMO_PURCHASE_FEATURE_LINES: tuple[tuple[str, float], ...] = (
    ("-Unlimited Play Time in three thrilling Game Modes!", 22.0),
    ("-The varied weapon arsenal consisting of over 20 unique", 17.0),
    (" weapons that allow you to deal death with plasma, lead,", 17.0),
    (" fire and electricity!", 22.0),
    ("-Over 40 game altering Perks!", 22.0),
    ("-40 insane Levels that give you", 18.0),
    (" hours of intense and fun gameplay!", 22.0),
    ("-The ability to post your high scores online!", 44.0),
)
_DEMO_PURCHASE_FOOTER = "Purchasing the game is very easy and secure."


class DemoState(Protocol):
    assets_dir: Path
    rng: random.Random
    config: CrimsonConfig
    texture_cache: PaqTextureCache | None
    audio: AudioState | None


def _weapon_name(weapon_id: int) -> str:
    for weapon in WEAPON_TABLE:
        if weapon.weapon_id == weapon_id:
            return weapon.name or f"weapon_{weapon_id}"
    return f"weapon_{weapon_id}"


def _clamp(value: float, lo: float, hi: float) -> float:
    if value < lo:
        return lo
    if value > hi:
        return hi
    return value


def _distance_sq(x0: float, y0: float, x1: float, y1: float) -> float:
    dx = x1 - x0
    dy = y1 - y0
    return dx * dx + dy * dy


def _normalize(dx: float, dy: float) -> tuple[float, float, float]:
    d = math.hypot(dx, dy)
    if d <= 1e-6:
        return 0.0, 0.0, 0.0
    inv = 1.0 / d
    return dx * inv, dy * inv, d


class DemoView:
    """Attract-mode demo scaffold.

    Modeled after the classic demo helpers in crimsonland.exe:
      - demo_setup_variant_0 @ 0x00402ED0
      - demo_setup_variant_1 @ 0x004030F0
      - demo_setup_variant_2 @ 0x00402FE0
      - demo_setup_variant_3 @ 0x00403250
      - demo_mode_start       @ 0x00403390
    """

    def __init__(self, state: DemoState) -> None:
        self._state = state
        self._world = GameWorld(
            assets_dir=state.assets_dir,
            world_size=WORLD_SIZE,
            demo_mode_active=True,
            hardcore=bool(int(state.config.data.get("hardcore_flag", 0) or 0)),
            difficulty_level=0,
            texture_cache=state.texture_cache,
            config=state.config,
            audio=state.audio,
            audio_rng=state.rng,
        )
        self._crand = Crand(0)
        self._demo_targets: list[int | None] = []
        self._variant_index = 0
        self._demo_variant_index = 0
        self._quest_spawn_timeline_ms = 0
        self._demo_time_limit_ms = 0
        self._finished = False
        self._upsell_message_index = 0
        self._upsell_pulse_ms = 0
        self._upsell_font: GrimMonoFont | None = None
        self._small_font: SmallFontData | None = None
        self._purchase_active = False
        self._purchase_url_opened = False
        self._spawn_rng = Crand(0)

    def open(self) -> None:
        self._finished = False
        self._upsell_message_index = 0
        self._upsell_pulse_ms = 0
        self._purchase_active = False
        self._purchase_url_opened = False
        self._variant_index = 0
        self._demo_variant_index = 0
        self._quest_spawn_timeline_ms = 0
        self._demo_time_limit_ms = 0
        self._crand.srand(self._state.rng.getrandbits(32))
        self._world.open()
        self._demo_mode_start()

    def close(self) -> None:
        self._world.close()
        if self._upsell_font is not None:
            rl.unload_texture(self._upsell_font.texture)
            self._upsell_font = None
        if self._small_font is not None:
            rl.unload_texture(self._small_font.texture)
            self._small_font = None

    def is_finished(self) -> bool:
        return self._finished

    def update(self, dt: float) -> None:
        if self._state.audio is not None:
            update_audio(self._state.audio, dt)
        if self._finished:
            return
        frame_dt = min(dt, 0.1)
        frame_dt_ms = int(frame_dt * 1000.0)
        if frame_dt_ms <= 0:
            return

        if (not self._purchase_active) and getattr(self._state, "demo_enabled", False) and self._purchase_screen_triggered():
            self._begin_purchase_screen(DEMO_PURCHASE_SCREEN_LIMIT_MS, reset_timeline=False)

        if self._purchase_active:
            self._upsell_pulse_ms += frame_dt_ms
            self._update_purchase_screen()
            self._quest_spawn_timeline_ms += frame_dt_ms
            if self._quest_spawn_timeline_ms > self._demo_time_limit_ms:
                # demo_purchase_screen_update restarts the demo once the purchase screen
                # timer exceeds demo_time_limit_ms.
                self._demo_mode_start()
            return

        if self._skip_triggered():
            self._finished = True
            return

        self._quest_spawn_timeline_ms += frame_dt_ms
        self._update_world(frame_dt)
        if self._quest_spawn_timeline_ms > self._demo_time_limit_ms:
            self._demo_mode_start()

    def draw(self) -> None:
        if self._purchase_active:
            self._draw_purchase_screen()
            return
        self._world.draw()
        self._draw_overlay()

    def _skip_triggered(self) -> bool:
        if rl.get_key_pressed() != 0:
            return True
        if rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
            return True
        if rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_RIGHT):
            return True
        return False

    def _purchase_screen_triggered(self) -> bool:
        if rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
            return True
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            return True
        if rl.is_key_pressed(rl.KeyboardKey.KEY_SPACE):
            return True
        return False

    def _begin_purchase_screen(self, limit_ms: int, *, reset_timeline: bool) -> None:
        self._purchase_active = True
        if reset_timeline:
            self._quest_spawn_timeline_ms = 0
        self._demo_time_limit_ms = max(0, int(limit_ms))
        self._purchase_url_opened = False

    def _ensure_small_font(self) -> SmallFontData:
        if self._small_font is not None:
            return self._small_font
        missing_assets: list[str] = []
        self._small_font = load_small_font(self._state.assets_dir, missing_assets)
        return self._small_font

    def _purchase_var_28_2(self) -> float:
        screen_w = int(self._state.config.screen_width)
        if screen_w == 0x320:  # 800
            return 64.0
        if screen_w == 0x400:  # 1024
            return 128.0
        return 0.0

    def _update_purchase_screen(self) -> None:
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._purchase_active = False
            self._finished = True
            return

        small = self._ensure_small_font()
        # ui_button_update uses the medium (145px wide) button sprite here (the per-button
        # "small" flag at +0x14 is 0 for both purchase/maybe-later globals).
        button_tex = self._ensure_cache().get_or_load("ui_button_md", "ui/ui_button_145x32.jaz").texture

        if button_tex is None:
            return

        w = float(self._state.config.screen_width)
        h = float(self._state.config.screen_height)
        wide_shift = self._purchase_var_28_2()
        button_x = w / 2.0 + 128.0
        button_base_y = h / 2.0 + 102.0 + wide_shift * 0.3
        purchase_y = button_base_y + 50.0
        maybe_y = button_base_y + 90.0

        purchase_rect = rl.Rectangle(button_x, purchase_y, float(button_tex.width), float(button_tex.height))
        maybe_rect = rl.Rectangle(button_x, maybe_y, float(button_tex.width), float(button_tex.height))

        mouse = rl.get_mouse_position()
        if (
            purchase_rect.x <= mouse.x <= purchase_rect.x + purchase_rect.width
            and purchase_rect.y <= mouse.y <= purchase_rect.y + purchase_rect.height
            and rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
        ):
            if not self._purchase_url_opened:
                self._purchase_url_opened = True
                try:
                    import webbrowser

                    webbrowser.open(DEMO_PURCHASE_URL)
                except Exception:
                    pass
            if hasattr(self._state, "quit_requested"):
                self._state.quit_requested = True

        if (
            maybe_rect.x <= mouse.x <= maybe_rect.x + maybe_rect.width
            and maybe_rect.y <= mouse.y <= maybe_rect.y + maybe_rect.height
            and rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
        ):
            self._purchase_active = False
            self._finished = True
            return

        # Keyboard activation for convenience; original uses UI mouse.
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ENTER):
            if not self._purchase_url_opened:
                self._purchase_url_opened = True
                try:
                    import webbrowser

                    webbrowser.open(DEMO_PURCHASE_URL)
                except Exception:
                    pass
            if hasattr(self._state, "quit_requested"):
                self._state.quit_requested = True

        # Keep small referenced to avoid unused warnings if this method grows.
        _ = small

    def _draw_purchase_screen(self) -> None:
        rl.clear_background(rl.BLACK)

        logos = getattr(self._state, "logos", None)
        if logos is None or logos.backplasma.texture is None:
            return
        backplasma = logos.backplasma.texture

        pulse_phase = float(self._upsell_pulse_ms % 1000)
        pulse = math.sin(pulse_phase * 6.2831855)
        pulse = pulse * pulse

        screen_w = float(self._state.config.screen_width)
        screen_h = float(self._state.config.screen_height)

        # demo_purchase_screen_update @ 0x0040b985:
        #   - full-screen quad
        #   - UV: 0..0.5 (top-left quarter of the backplasma atlas)
        #   - per-corner color slots, with a sin^2 pulse at bottom-right

        def _to_u8(value: float) -> int:
            return int(_clamp(value, 0.0, 1.0) * 255.0 + 0.5)

        c0 = rl.Color(_to_u8(0.0), _to_u8(0.0), _to_u8(0.0), _to_u8(1.0))
        c1 = rl.Color(_to_u8(0.0), _to_u8(0.0), _to_u8(0.3), _to_u8(1.0))
        c2 = rl.Color(
            _to_u8(0.0),
            _to_u8(0.4),
            _to_u8(pulse * 0.55),
            _to_u8(pulse),
        )
        c3 = rl.Color(_to_u8(0.0), _to_u8(0.4), _to_u8(0.4), _to_u8(1.0))

        rl.begin_blend_mode(rl.BLEND_ALPHA)
        rl.rl_set_texture(backplasma.id)
        rl.rl_begin(rl.RL_QUADS)
        # TL
        rl.rl_color4ub(c0.r, c0.g, c0.b, c0.a)
        rl.rl_tex_coord2f(0.0, 0.0)
        rl.rl_vertex2f(0.0, 0.0)
        # TR
        rl.rl_color4ub(c1.r, c1.g, c1.b, c1.a)
        rl.rl_tex_coord2f(0.5, 0.0)
        rl.rl_vertex2f(screen_w, 0.0)
        # BR
        rl.rl_color4ub(c2.r, c2.g, c2.b, c2.a)
        rl.rl_tex_coord2f(0.5, 0.5)
        rl.rl_vertex2f(screen_w, screen_h)
        # BL
        rl.rl_color4ub(c3.r, c3.g, c3.b, c3.a)
        rl.rl_tex_coord2f(0.0, 0.5)
        rl.rl_vertex2f(0.0, screen_h)
        rl.rl_end()
        rl.rl_set_texture(0)
        rl.end_blend_mode()

        wide_shift = self._purchase_var_28_2()

        # Mockup and logo textures.
        if logos.mockup.texture is not None:
            mockup = logos.mockup.texture
            x = screen_w / 2.0 - 128.0 + wide_shift
            y = screen_h / 2.0 - 140.0
            dst = rl.Rectangle(x, y, 512.0, 256.0)
            src = rl.Rectangle(0.0, 0.0, float(mockup.width), float(mockup.height))
            rl.draw_texture_pro(mockup, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)

        if logos.cl_logo.texture is not None:
            cl_logo = logos.cl_logo.texture
            x = screen_w / 2.0 - 256.0
            y = screen_h / 2.0 - 200.0 - wide_shift * 0.4
            dst = rl.Rectangle(x, y, 512.0, 64.0)
            src = rl.Rectangle(0.0, 0.0, float(cl_logo.width), float(cl_logo.height))
            rl.draw_texture_pro(cl_logo, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)

        small = self._ensure_small_font()
        text_scale = 1.2
        x_text = screen_w / 2.0 - 296.0 - wide_shift * 0.8
        y = screen_h / 2.0 - 104.0
        color = rl.Color(255, 255, 255, 255)
        draw_small_text(small, _DEMO_PURCHASE_TITLE, x_text, y, text_scale, color)
        y += 28.0
        draw_small_text(small, _DEMO_PURCHASE_FEATURES_TITLE, x_text, y, text_scale, color)

        underline_w = measure_small_text_width(small, _DEMO_PURCHASE_FEATURES_TITLE, text_scale)
        rl.draw_rectangle_rec(rl.Rectangle(x_text, y + 15.0, underline_w, 2.0), rl.Color(255, 255, 255, 160))

        y += 22.0
        x_list = x_text + 8.0
        for line, delta_y in _DEMO_PURCHASE_FEATURE_LINES:
            draw_small_text(small, line, x_list, y, text_scale, color)
            y += delta_y
        draw_small_text(small, _DEMO_PURCHASE_FOOTER, x_text, y, text_scale, color)

        # Buttons on the right.
        cache = self._ensure_cache()
        button_tex = cache.get_or_load("ui_button_md", "ui/ui_button_145x32.jaz").texture
        if button_tex is None:
            return

        button_x = screen_w / 2.0 + 128.0
        button_base_y = screen_h / 2.0 + 102.0 + wide_shift * 0.3
        purchase_y = button_base_y + 50.0
        maybe_y = button_base_y + 90.0
        mouse = rl.get_mouse_position()

        def draw_button(texture: rl.Texture2D, label: str, x: float, y0: float) -> None:
            hovered = x <= mouse.x <= x + texture.width and y0 <= mouse.y <= y0 + texture.height
            tint = rl.Color(255, 255, 255, 255) if hovered else rl.Color(220, 220, 220, 255)
            rl.draw_texture(texture, int(x), int(y0), tint)
            label_scale = 1.0
            text_w = measure_small_text_width(small, label, label_scale)
            text_x = x + float(texture.width) * 0.5 - text_w * 0.5 + 1.0
            text_y = y0 + 10.0
            alpha = 1.0 if hovered else 0.7
            draw_small_text(small, label, text_x, text_y, label_scale, rl.Color(255, 255, 255, int(255 * alpha)))

        draw_button(button_tex, "Purchase", button_x, purchase_y)
        draw_button(button_tex, "Maybe later", button_x, maybe_y)

        # Demo purchase screen uses menu-style cursor; draw it explicitly since the OS cursor is hidden.
        particles = cache.get_or_load("particles", "game/particles.jaz").texture
        cursor_tex = cache.get_or_load("ui_cursor", "ui/ui_cursor.jaz").texture
        pulse_time = float(self._upsell_pulse_ms) * 0.001
        draw_menu_cursor(particles, cursor_tex, x=float(mouse.x), y=float(mouse.y), pulse_time=pulse_time)

    def _ensure_cache(self) -> PaqTextureCache:
        cache = self._state.texture_cache
        if cache is not None:
            return cache
        entries = load_paq_entries(self._state.assets_dir)
        cache = PaqTextureCache(entries=entries, textures={})
        self._state.texture_cache = cache
        return cache

    def _demo_mode_start(self) -> None:
        index = self._demo_variant_index
        self._demo_variant_index = (index + 1) % DEMO_VARIANT_COUNT
        self._variant_index = index
        self._quest_spawn_timeline_ms = 0
        self._demo_time_limit_ms = 0
        self._purchase_active = False
        self._purchase_url_opened = False
        self._spawn_rng.srand(self._state.rng.randrange(0, 0x1_0000_0000))
        self._world.state.bonuses.weapon_power_up = 0.0
        if index == 0:
            self._apply_variant_ground(0)
            self._setup_variant_0()
        elif index == 1:
            self._apply_variant_ground(1)
            self._setup_variant_1()
        elif index == 2:
            self._apply_variant_ground(2)
            self._setup_variant_2()
        elif index == 3:
            self._apply_variant_ground(3)
            self._setup_variant_3()
        elif index == 4:
            self._apply_variant_ground(4)
            self._setup_variant_0()
        else:
            # demo_purchase_interstitial_begin
            self._begin_purchase_screen(DEMO_PURCHASE_INTERSTITIAL_LIMIT_MS, reset_timeline=True)

        # demo_purchase_screen_update increments demo_upsell_message_index when the
        # timeline resets (quest_spawn_timeline == 0) and the purchase screen is inactive.
        if (not self._purchase_active) and _DEMO_UPSELL_MESSAGES:
            self._upsell_message_index = (self._upsell_message_index + 1) % len(_DEMO_UPSELL_MESSAGES)

    def _setup_world_players(self, specs: list[tuple[float, float, int]]) -> None:
        seed = int(self._state.rng.getrandbits(32))
        self._world.reset(seed=seed, player_count=len(specs))
        for idx, (x, y, weapon_id) in enumerate(specs):
            if idx >= len(self._world.players):
                continue
            player = self._world.players[idx]
            player.pos_x = float(x)
            player.pos_y = float(y)
            # Keep aim anchored to the spawn position so demo aim starts stable.
            player.aim_x = float(x)
            player.aim_y = float(y)
            weapon_assign_player(player, int(weapon_id))
        self._demo_targets = [None] * len(self._world.players)

    def _apply_variant_ground(self, index: int) -> None:
        if index == 5:
            return
        terrain = {
            0: (
                "ter_q1_base",
                "ter_q1_tex1",
                "ter/ter_q1_base.jaz",
                "ter/ter_q1_tex1.jaz",
            ),
            1: (
                "ter_q2_base",
                "ter_q2_tex1",
                "ter/ter_q2_base.jaz",
                "ter/ter_q2_tex1.jaz",
            ),
            2: (
                "ter_q3_base",
                "ter_q3_tex1",
                "ter/ter_q3_base.jaz",
                "ter/ter_q3_tex1.jaz",
            ),
            3: (
                "ter_q4_base",
                "ter_q4_tex1",
                "ter/ter_q4_base.jaz",
                "ter/ter_q4_tex1.jaz",
            ),
            4: (
                "ter_q1_base",
                "ter_q1_tex1",
                "ter/ter_q1_base.jaz",
                "ter/ter_q1_tex1.jaz",
            ),
        }.get(
            index,
            (
                "ter_q1_base",
                "ter_q1_tex1",
                "ter/ter_q1_base.jaz",
                "ter/ter_q1_tex1.jaz",
            ),
        )
        base_key, overlay_key, base_path, overlay_path = terrain
        self._world.set_terrain(
            base_key=base_key,
            overlay_key=overlay_key,
            base_path=base_path,
            overlay_path=overlay_path,
        )

    def _wrap_pos(self, x: float, y: float) -> tuple[float, float]:
        return (x % WORLD_SIZE, y % WORLD_SIZE)

    def _crand_mod(self, mod: int) -> int:
        if mod <= 0:
            return 0
        return int(self._crand.rand() % mod)

    def _spawn(self, spawn_id: int, x: float, y: float, *, heading: float = 0.0) -> None:
        x, y = self._wrap_pos(x, y)
        self._world.creatures.spawn_template(
            int(spawn_id),
            (x, y),
            float(heading),
            self._spawn_rng,
            rand=self._spawn_rng.rand,
        )

    def _setup_variant_0(self) -> None:
        self._demo_time_limit_ms = 4000
        # demo_setup_variant_0 uses weapon_id=0x0B.
        weapon_id = 11
        self._setup_world_players(
            [
                (448.0, 384.0, weapon_id),
                (546.0, 654.0, weapon_id),
            ]
        )
        y = 256
        i = 0
        while y < 1696:
            col = i % 2
            self._spawn(0x38, float((col + 2) * 64), float(y), heading=-100.0)
            self._spawn(0x38, float(col * 64 + 798), float(y), heading=-100.0)
            y += 80
            i += 1

    def _setup_variant_1(self) -> None:
        self._demo_time_limit_ms = 5000
        # demo_setup_variant_1 uses weapon_id=0x05.
        weapon_id = 5
        self._setup_world_players(
            [
                (490.0, 448.0, weapon_id),
                (480.0, 576.0, weapon_id),
            ]
        )
        self._world.state.bonuses.weapon_power_up = 15.0
        for idx in range(20):
            x = float(self._crand_mod(200) + 32)
            y = float(self._crand_mod(899) + 64)
            self._spawn(0x34, x, y, heading=-100.0)
            if idx % 3 != 0:
                x2 = float(self._crand_mod(30) + 32)
                y2 = float(self._crand_mod(899) + 64)
                self._spawn(0x35, x2, y2, heading=-100.0)

    def _setup_variant_2(self) -> None:
        self._demo_time_limit_ms = 5000
        # demo_setup_variant_2 uses weapon_id=0x15.
        weapon_id = 21
        self._setup_world_players([(512.0, 512.0, weapon_id)])
        y = 128
        i = 0
        while y < 848:
            col = i % 2
            self._spawn(0x41, float(col * 64 + 32), float(y), heading=-100.0)
            self._spawn(0x41, float((col + 2) * 64), float(y), heading=-100.0)
            self._spawn(0x41, float(col * 64 - 64), float(y), heading=-100.0)
            self._spawn(0x41, float((col + 12) * 64), float(y), heading=-100.0)
            y += 60
            i += 1

    def _setup_variant_3(self) -> None:
        self._demo_time_limit_ms = 4000
        # demo_setup_variant_3 uses weapon_id=0x12.
        weapon_id = 18
        self._setup_world_players([(512.0, 512.0, weapon_id)])
        for idx in range(20):
            x = float(self._crand_mod(200) + 32)
            y = float(self._crand_mod(899) + 64)
            self._spawn(0x24, x, y, heading=0.0)
            if idx % 3 != 0:
                x2 = float(self._crand_mod(30) + 32)
                y2 = float(self._crand_mod(899) + 64)
                self._spawn(0x25, x2, y2, heading=0.0)

    def _world_params(self) -> tuple[float, float, float, float]:
        out_w = float(rl.get_screen_width())
        out_h = float(rl.get_screen_height())
        screen_w = float(self._state.config.screen_width)
        screen_h = float(self._state.config.screen_height)
        if screen_w > WORLD_SIZE:
            screen_w = WORLD_SIZE
        if screen_h > WORLD_SIZE:
            screen_h = WORLD_SIZE

        cam_x = self._camera_x
        cam_y = self._camera_y
        min_x = screen_w - WORLD_SIZE
        min_y = screen_h - WORLD_SIZE
        if cam_x > -1.0:
            cam_x = -1.0
        if cam_y > -1.0:
            cam_y = -1.0
        if cam_x < min_x:
            cam_x = min_x
        if cam_y < min_y:
            cam_y = min_y

        scale_x = out_w / screen_w if screen_w > 0 else 1.0
        scale_y = out_h / screen_h if screen_h > 0 else 1.0
        return cam_x, cam_y, scale_x, scale_y

    def _world_to_screen(self, x: float, y: float) -> tuple[float, float]:
        cam_x, cam_y, scale_x, scale_y = self._world_params()
        return (x + cam_x) * scale_x, (y + cam_y) * scale_y

    def _select_frame(self, spawn_id: int, phase: float) -> tuple[int, bool]:
        template = SPAWN_ID_TO_TEMPLATE.get(spawn_id)
        if template is None or template.type_id is None:
            return 0, False
        info = _TYPE_ANIM.get(template.type_id)
        if info is None:
            return 0, False
        flags = template.flags or CreatureFlags(0)
        frame, mirror_applied, _ = creature_anim_select_frame(
            phase,
            base_frame=info.base,
            mirror_long=info.mirror,
            flags=flags,
        )
        return frame, mirror_applied

    def _draw_fx(self) -> None:
        projectiles = self._projectile_pool.iter_active()
        secondary = self._secondary_projectile_pool.iter_active()
        if not (projectiles or secondary or self._beams or self._explosions):
            return
        cam_x, cam_y, scale_x, scale_y = self._world_params()
        del cam_x, cam_y
        scale = (scale_x + scale_y) * 0.5

        for proj in projectiles:
            sx, sy = self._world_to_screen(proj.pos_x, proj.pos_y)
            base_radius = {
                0x05: 6.0,  # gauss
                0x0B: 10.0,  # rocket launcher
                0x15: 7.0,  # ion minigun beam seed
            }.get(proj.type_id, 5.0)
            radius = max(1.0, base_radius * scale)
            color = {
                0x05: rl.Color(235, 235, 235, 255),
                0x0B: rl.Color(255, 120, 80, 255),
                0x15: rl.Color(120, 220, 255, 255),
            }.get(proj.type_id, rl.Color(235, 235, 235, 255))
            rl.draw_circle(int(sx), int(sy), radius, color)

        for proj in secondary:
            sx, sy = self._world_to_screen(proj.pos_x, proj.pos_y)
            if proj.type_id == 4:
                radius = max(1.0, 12.0 * scale)
                rl.draw_circle(int(sx), int(sy), radius, rl.Color(200, 120, 255, 255))
                continue
            if proj.type_id == 3:
                t = _clamp(proj.vel_x, 0.0, 1.0)
                radius = proj.vel_y * t * 80.0
                alpha = int((1.0 - t) * 180.0)
                color = rl.Color(200, 120, 255, alpha)
                rl.draw_circle_lines(int(sx), int(sy), max(1.0, radius * scale), color)

        for beam in self._beams:
            x0, y0 = self._world_to_screen(beam.x0, beam.y0)
            x1, y1 = self._world_to_screen(beam.x1, beam.y1)
            alpha = int(_clamp(beam.life / 0.08, 0.0, 1.0) * 255.0)
            color = rl.Color(120, 220, 255, alpha)
            rl.draw_line_ex(rl.Vector2(x0, y0), rl.Vector2(x1, y1), 2.0 * scale, color)

        for fx in self._explosions:
            t = fx.elapsed / fx.duration if fx.duration > 0 else 1.0
            radius = fx.max_radius * _clamp(t, 0.0, 1.0)
            sx, sy = self._world_to_screen(fx.x, fx.y)
            alpha = int((1.0 - _clamp(t, 0.0, 1.0)) * 180.0)
            color = rl.Color(255, 180, 100, alpha) if fx.kind == "rocket" else rl.Color(200, 120, 255, alpha)
            rl.draw_circle_lines(int(sx), int(sy), max(1.0, radius * scale), color)

    def _draw_entities(self) -> None:
        cache = self._state.texture_cache
        if cache is None:
            return
        cam_x, cam_y, scale_x, scale_y = self._world_params()
        del cam_x, cam_y

        player_tex = cache.get_or_load("trooper", "game/trooper.jaz").texture
        if player_tex is not None:
            for player in self._players:
                self._draw_sprite(
                    player_tex,
                    CreatureTypeId.TROOPER,
                    CreatureFlags(0),
                    player.phase,
                    player.x,
                    player.y,
                    scale_x,
                    scale_y,
                    tint=rl.Color(240, 240, 255, 255),
                )

        for creature in self._creatures:
            type_id = creature.type_id
            if type_id is None:
                continue
            asset = _TYPE_ASSET.get(type_id)
            if asset is None:
                continue
            texture = cache.texture(asset)
            if texture is None:
                rel_path = f"game/{asset}.jaz"
                texture = cache.get_or_load(asset, rel_path).texture
            if texture is None:
                continue
            flags = creature.flags

            def _to_u8(value: float) -> int:
                return int(_clamp(value, 0.0, 1.0) * 255.0 + 0.5)

            tint = rl.WHITE
            if creature.tint is not None and any(v is not None for v in creature.tint):
                tint_r, tint_g, tint_b, tint_a = resolve_tint(creature.tint)
                tint = rl.Color(
                    _to_u8(tint_r),
                    _to_u8(tint_g),
                    _to_u8(tint_b),
                    _to_u8(tint_a),
                )
            self._draw_sprite(
                texture,
                type_id,
                flags,
                creature.anim_phase,
                creature.x,
                creature.y,
                scale_x,
                scale_y,
                tint=tint,
                size_scale=_clamp(creature.size / 64.0, 0.25, 2.0),
            )

    def _draw_sprite(
        self,
        texture: rl.Texture2D,
        type_id: CreatureTypeId,
        flags: CreatureFlags,
        phase: float,
        world_x: float,
        world_y: float,
        scale_x: float,
        scale_y: float,
        *,
        tint: rl.Color,
        size_scale: float = 1.0,
    ) -> None:
        info = _TYPE_ANIM.get(type_id)
        if info is None:
            return
        frame, _, _ = creature_anim_select_frame(
            phase,
            base_frame=info.base,
            mirror_long=info.mirror,
            flags=flags,
        )

        grid = 8
        cell = float(texture.width) / grid if grid > 0 else float(texture.width)
        row = frame // grid
        col = frame % grid
        src = rl.Rectangle(float(col * cell), float(row * cell), float(cell), float(cell))
        screen_x, screen_y = self._world_to_screen(world_x, world_y)
        width = cell * scale_x * size_scale
        height = cell * scale_y * size_scale
        dst = rl.Rectangle(screen_x, screen_y, width, height)
        origin = rl.Vector2(width * 0.5, height * 0.5)
        rl.draw_texture_pro(texture, src, dst, origin, 0.0, tint)

    def _draw_overlay(self) -> None:
        if getattr(self._state, "demo_enabled", False):
            self._draw_demo_upsell_overlay()
            return
        title = f"DEMO MODE  ({self._variant_index + 1}/{DEMO_VARIANT_COUNT})"
        hint = "Press any key / click to skip"
        remaining = max(0.0, float(self._demo_time_limit_ms - self._quest_spawn_timeline_ms) / 1000.0)
        weapons = ", ".join(f"P{p.index + 1}:{_weapon_name(p.weapon_id)}" for p in self._world.players)
        detail = f"{weapons}  —  next in {remaining:0.1f}s"
        rl.draw_text(title, 16, 12, 20, rl.Color(240, 240, 240, 255))
        rl.draw_text(detail, 16, 36, 16, rl.Color(180, 180, 190, 255))
        rl.draw_text(hint, 16, 56, 16, rl.Color(140, 140, 150, 255))

    def _ensure_upsell_font(self) -> GrimMonoFont:
        if self._upsell_font is not None:
            return self._upsell_font
        missing_assets: list[str] = []
        self._upsell_font = load_grim_mono_font(self._state.assets_dir, missing_assets)
        return self._upsell_font

    def _draw_demo_upsell_overlay(self) -> None:
        # Modeled after the shareware "Want more ..." overlay in demo_purchase_screen_update
        # (crimsonland.exe 0x0040B740), but without the purchase screen.
        if not _DEMO_UPSELL_MESSAGES:
            return

        font = self._ensure_upsell_font()
        msg = _DEMO_UPSELL_MESSAGES[self._upsell_message_index]

        timeline_ms = self._quest_spawn_timeline_ms
        limit_ms = self._demo_time_limit_ms
        var_2c = float(timeline_ms) * 0.016

        alpha = 1.0
        if var_2c < 20.0:
            alpha = var_2c * 0.05
        if timeline_ms > limit_ms - 500:
            alpha = float(limit_ms - timeline_ms) * 0.002
        alpha = _clamp(alpha, 0.0, 1.0)

        scale = 0.8
        text_w = float(len(msg)) * 12.8

        text_x = 50.0
        text_y = var_2c + 50.0
        bg_x = 60.0
        bg_y = text_y - 4.0
        bar_x = 64.0
        bar_y = var_2c + 72.0

        bg_alpha = int(round(_clamp(alpha * 0.5, 0.0, 1.0) * 255.0))
        bar_alpha = int(round(_clamp(alpha * 0.8, 0.0, 1.0) * 255.0))
        txt_alpha = int(round(_clamp(alpha, 0.0, 1.0) * 255.0))

        rl.draw_rectangle_rec(
            rl.Rectangle(bg_x, bg_y, text_w + 12.0, 30.0),
            rl.Color(0, 0, 0, bg_alpha),
        )

        progress = 0.0
        if limit_ms > 0:
            progress = _clamp(float(timeline_ms) / float(limit_ms), 0.0, 1.0)
        rl.draw_rectangle_rec(
            rl.Rectangle(bar_x, bar_y, text_w * progress, 3.0),
            rl.Color(128, 26, 26, bar_alpha),
        )

        draw_grim_mono_text(font, msg, text_x, text_y, scale, rl.Color(255, 255, 255, txt_alpha))

    def _update_world(self, dt: float) -> None:
        if not self._world.players:
            return
        inputs = self._build_demo_inputs(dt)
        self._world.update(dt, inputs=inputs, auto_pick_perks=False, game_mode=0, perk_progression_enabled=False)

    def _build_demo_inputs(self, dt: float) -> list[PlayerInput]:
        players = self._world.players
        creatures = self._world.creatures.entries
        if len(self._demo_targets) != len(players):
            self._demo_targets = [None] * len(players)
        center_x = float(self._world.world_size) * 0.5
        center_y = float(self._world.world_size) * 0.5

        dt = float(dt)

        def _turn_towards_heading(cur: float, target: float) -> tuple[float, float]:
            cur = cur % math.tau
            target = target % math.tau
            delta = (target - cur + math.pi) % math.tau - math.pi
            diff = abs(delta)
            if diff <= 1e-9:
                return cur, 0.0
            step = dt * diff * 5.0
            cur = (cur + step) % math.tau if delta > 0.0 else (cur - step) % math.tau
            return cur, diff

        inputs: list[PlayerInput] = []
        for idx, player in enumerate(players):
            target_idx = self._select_demo_target(idx, player, creatures)
            target = None
            if target_idx is not None and 0 <= target_idx < len(creatures):
                candidate = creatures[target_idx]
                if candidate.active and candidate.hp > 0.0:
                    target = candidate

            # Aim: ease the aim point toward the target.
            aim_x = float(player.aim_x)
            aim_y = float(player.aim_y)
            auto_fire = False
            if target is not None:
                aim_dx = float(target.x) - aim_x
                aim_dy = float(target.y) - aim_y
                aim_dir_x, aim_dir_y, aim_dist = _normalize(aim_dx, aim_dy)
                if aim_dist >= 4.0:
                    step = aim_dist * 6.0 * dt
                    aim_x += aim_dir_x * step
                    aim_y += aim_dir_y * step
                else:
                    aim_x = float(target.x)
                    aim_y = float(target.y)
                auto_fire = aim_dist < 128.0
            else:
                ax, ay, amag = _normalize(float(player.pos_x) - center_x, float(player.pos_y) - center_y)
                if amag <= 1e-6:
                    ax, ay = 0.0, -1.0
                aim_x = float(player.pos_x) + ax * 60.0
                aim_y = float(player.pos_y) + ay * 60.0

            # Movement:
            # - orbit center if no target
            # - chase target when near center
            # - return to center when too far
            if target is None:
                move_dx = -(float(player.pos_y) - center_y)
                move_dy = float(player.pos_x) - center_x
            else:
                center_dist = math.hypot(float(player.pos_x) - center_x, float(player.pos_y) - center_y)
                if center_dist <= 300.0:
                    move_dx = float(target.x) - float(player.pos_x)
                    move_dy = float(target.y) - float(player.pos_y)
                else:
                    move_dx = center_x - float(player.pos_x)
                    move_dy = center_y - float(player.pos_y)

            desired_x, desired_y, desired_mag = _normalize(move_dx, move_dy)
            if desired_mag <= 1e-6:
                move_x = 0.0
                move_y = 0.0
            else:
                desired_heading = math.atan2(desired_y, desired_x) + math.pi / 2.0
                smoothed_heading, angle_diff = _turn_towards_heading(float(player.heading), desired_heading)
                move_mag = max(0.001, (math.pi - angle_diff) / math.pi)
                move_x = math.cos(smoothed_heading - math.pi / 2.0) * move_mag
                move_y = math.sin(smoothed_heading - math.pi / 2.0) * move_mag

            inputs.append(
                PlayerInput(
                    move_x=move_x,
                    move_y=move_y,
                    aim_x=aim_x,
                    aim_y=aim_y,
                    fire_down=auto_fire,
                    fire_pressed=auto_fire,
                    reload_pressed=False,
                )
            )

        return inputs

    def _nearest_world_creature_index(self, x: float, y: float) -> int | None:
        best_idx = None
        best_dist = 0.0
        for idx, creature in enumerate(self._world.creatures.entries):
            if not (creature.active and creature.hp > 0.0):
                continue
            d = _distance_sq(x, y, creature.x, creature.y)
            if best_idx is None or d < best_dist:
                best_idx = idx
                best_dist = d
        return best_idx

    def _select_demo_target(self, player_index: int, player: PlayerState, creatures: list) -> int | None:
        candidate = self._nearest_world_creature_index(player.pos_x, player.pos_y)
        current = self._demo_targets[player_index] if player_index < len(self._demo_targets) else None
        if current is None:
            self._demo_targets[player_index] = candidate
            return candidate
        if not (0 <= current < len(creatures)):
            self._demo_targets[player_index] = candidate
            return candidate
        current_creature = creatures[current]
        if current_creature.hp <= 0.0 or not current_creature.active:
            self._demo_targets[player_index] = candidate
            return candidate
        if candidate is None or candidate == current:
            return current
        cand_creature = creatures[candidate]
        if not cand_creature.active or cand_creature.hp <= 0.0:
            return current
        cur_d = math.hypot(current_creature.x - player.pos_x, current_creature.y - player.pos_y)
        cand_d = math.hypot(cand_creature.x - player.pos_x, cand_creature.y - player.pos_y)
        if cand_d + 64.0 < cur_d:
            self._demo_targets[player_index] = candidate
            return candidate
        return current

    def _update_sim(self, dt: float, dt_ms: int) -> None:
        self._bonus_weapon_power_up_timer = max(0.0, self._bonus_weapon_power_up_timer - dt)
        self._update_creatures(dt, dt_ms)
        self._update_spawn_slots(dt)
        self._update_projectiles(dt)
        self._update_players(dt)
        self._update_fx(dt)
        self._update_camera(dt)

    def _nearest_player_index(self, x: float, y: float) -> int | None:
        best_idx = None
        best_dist = 0.0
        for idx, player in enumerate(self._players):
            d = _distance_sq(x, y, player.x, player.y)
            if best_idx is None or d < best_dist:
                best_idx = idx
                best_dist = d
        return best_idx

    def _nearest_creature_index(self, x: float, y: float) -> int | None:
        best_idx = None
        best_dist = 0.0
        for idx, creature in enumerate(self._creatures):
            if creature.hp <= 0.0:
                continue
            d = _distance_sq(x, y, creature.x, creature.y)
            if best_idx is None or d < best_dist:
                best_idx = idx
                best_dist = d
        return best_idx

    def _update_spawn_slots(self, dt: float) -> None:
        if not self._spawn_slots:
            return

        spawn_events: list[tuple[int, float, float]] = []
        slot_count = len(self._spawn_slots)
        for slot_idx in range(slot_count):
            slot = self._spawn_slots[slot_idx]
            owner_idx = slot.owner_creature
            if not (0 <= owner_idx < len(self._creatures)):
                continue
            owner = self._creatures[owner_idx]
            if owner.hp <= 0.0:
                continue
            child_template_id = tick_spawn_slot(slot, dt)
            if child_template_id is None:
                continue
            spawn_events.append((child_template_id, owner.x, owner.y))

        for child_template_id, x, y in spawn_events:
            self._spawn(child_template_id, x, y, heading=-100.0)

    def _update_creatures(self, dt: float, dt_ms: int) -> None:
        if not self._creatures or not self._players:
            return
        for creature in self._creatures:
            if creature.hp <= 0.0:
                continue
            type_id = creature.type_id
            if type_id is None:
                template = SPAWN_ID_TO_TEMPLATE.get(creature.spawn_id)
                type_id = template.type_id if template is not None else None

            move_speed = creature.move_speed
            if move_speed <= 0.0:
                move_speed = self._creature_speed(type_id) / 30.0

            creature_ai7_tick_link_timer(creature, dt_ms=dt_ms, rand=self._crand.rand)

            target_idx = self._nearest_player_index(creature.x, creature.y)
            creature.target_player = target_idx
            if target_idx is None:
                creature.vx = 0.0
                creature.vy = 0.0
                continue
            target = self._players[target_idx]
            ai = creature_ai_update_target(
                creature,
                player_x=target.x,
                player_y=target.y,
                creatures=self._creatures,
                dt=dt,
            )
            creature.move_scale = ai.move_scale
            if ai.self_damage is not None:
                creature.hp -= ai.self_damage
                if creature.hp <= 0.0:
                    continue

            if creature.ai_mode == 7:
                creature.vx = 0.0
                creature.vy = 0.0
                continue

            creature.heading = _angle_approach(
                creature.heading, creature.target_heading, move_speed * 0.33333334 * 4.0, dt
            )
            speed = move_speed * 30.0
            direction_x = math.cos(creature.heading - math.pi / 2.0)
            direction_y = math.sin(creature.heading - math.pi / 2.0)
            creature.vx = direction_x * dt * creature.move_scale * speed
            creature.vy = direction_y * dt * creature.move_scale * speed

            radius = max(0.0, creature.size)
            creature.x = _clamp(creature.x + creature.vx, radius, WORLD_SIZE - radius)
            creature.y = _clamp(creature.y + creature.vy, radius, WORLD_SIZE - radius)

    def _select_player_target(self, player: DemoPlayer) -> int | None:
        candidate = self._nearest_creature_index(player.x, player.y)
        current = player.target_creature
        if current is None:
            return candidate
        if not (0 <= current < len(self._creatures)):
            return candidate
        current_creature = self._creatures[current]
        if current_creature.hp <= 0.0:
            return candidate
        if candidate is None or candidate == current:
            return current
        cand_creature = self._creatures[candidate]
        if cand_creature.hp <= 0.0:
            return current
        cur_d = math.hypot(current_creature.x - player.x, current_creature.y - player.y)
        cand_d = math.hypot(cand_creature.x - player.x, cand_creature.y - player.y)
        if cand_d + 64.0 < cur_d:
            return candidate
        return current

    def _update_players(self, dt: float) -> None:
        if not self._players:
            return
        center_x = WORLD_SIZE * 0.5
        center_y = WORLD_SIZE * 0.5
        shot_cooldown_decay = dt * (1.5 if self._bonus_weapon_power_up_timer > 0.0 else 1.0)
        for player in self._players:
            player.shot_cooldown = max(0.0, player.shot_cooldown - shot_cooldown_decay)
            player.spread_heat = max(0.01, player.spread_heat - dt * 0.4)

            if player.reload_timer > 0.0:
                player.reload_timer = max(0.0, player.reload_timer - dt)
                if player.reload_timer <= 0.0:
                    weapon = self._weapon_entry(player.weapon_id)
                    clip_size = int(weapon.clip_size) if weapon is not None and weapon.clip_size is not None else 0
                    player.ammo = max(0, clip_size)
                    player.reload_timer = 0.0
                    player.reload_timer_max = 0.0

            player.target_creature = self._select_player_target(player)
            target = self._creatures[player.target_creature] if player.target_creature is not None else None
            if target is not None and target.hp > 0.0:
                dx = target.x - player.x
                dy = target.y - player.y
                nx, ny, _ = _normalize(dx, dy)
                player.aim_x, player.aim_y = nx, ny
            else:
                dx = center_x - player.x
                dy = center_y - player.y
                nx, ny, _ = _normalize(dx, dy)
                player.aim_x, player.aim_y = nx, ny

            move_x, move_y = 0.0, 0.0
            to_cx = center_x - player.x
            to_cy = center_y - player.y
            nx, ny, d = _normalize(to_cx, to_cy)
            if d > 120.0:
                move_x += nx
                move_y += ny

            if target is not None and target.hp > 0.0:
                rx = player.x - target.x
                ry = player.y - target.y
                rnx, rny, rd = _normalize(rx, ry)
                if 0.0 < rd < 160.0:
                    strength = (160.0 - rd) / 160.0
                    move_x += rnx * (1.5 * strength)
                    move_y += rny * (1.5 * strength)

            orbit_dir = -1.0 if (player.index % 2) else 1.0
            ox, oy, _ = _normalize(-(player.y - center_y), player.x - center_x)
            move_x += ox * 0.55 * orbit_dir
            move_y += oy * 0.55 * orbit_dir

            mnx, mny, _ = _normalize(move_x, move_y)
            speed = 150.0
            player.vx = mnx * speed
            player.vy = mny * speed
            player.x = _clamp(player.x + player.vx * dt, 0.0, WORLD_SIZE)
            player.y = _clamp(player.y + player.vy * dt, 0.0, WORLD_SIZE)

            self._player_fire(player, target)

    def _player_fire(self, player: DemoPlayer, target: DemoCreature | None) -> None:
        weapon = self._weapon_entry(player.weapon_id)
        if weapon is None:
            return

        if player.reload_timer > 0.0:
            return
        if player.shot_cooldown > 0.0:
            return
        if target is None or target.hp <= 0.0:
            return

        if player.ammo <= 0:
            reload_time = float(weapon.reload_time) if weapon.reload_time is not None else 0.0
            if self._bonus_weapon_power_up_timer > 0.0:
                reload_time *= 0.6
            player.reload_timer_max = max(0.0, reload_time)
            player.reload_timer = player.reload_timer_max
            play_sfx(self._state.audio, resolve_weapon_sfx_ref(weapon.reload_sound), rng=self._state.rng)
            return

        shot_cooldown = float(weapon.shot_cooldown) if weapon.shot_cooldown is not None else 0.0
        player.shot_cooldown = max(0.02, shot_cooldown)

        spread_inc = float(weapon.spread_heat_inc) if weapon.spread_heat_inc is not None else 0.0
        player.spread_heat = min(0.48, max(0.01, player.spread_heat + spread_inc))

        theta = math.atan2(player.aim_y, player.aim_x)
        if player.spread_heat > 0.0:
            theta += (self._crand_float01() * 2.0 - 1.0) * player.spread_heat
        angle = theta + math.pi / 2.0

        muzzle_x = player.x + player.aim_x * 16.0
        muzzle_y = player.y + player.aim_y * 16.0

        play_sfx(self._state.audio, resolve_weapon_sfx_ref(weapon.fire_sound), rng=self._state.rng)

        if player.weapon_id in {WeaponId.GAUSS_GUN, WeaponId.ION_MINIGUN}:
            meta = float(weapon.projectile_meta) if weapon.projectile_meta is not None else 0.0
            if meta <= 0.0:
                meta = 45.0
            type_id = projectile_type_id_from_weapon_id(player.weapon_id)
            if type_id is None:
                return
            self._projectile_pool.spawn(
                pos_x=muzzle_x,
                pos_y=muzzle_y,
                angle=angle,
                type_id=type_id,
                owner_id=-100,
                base_damage=meta,
            )
        elif player.weapon_id == WeaponId.ROCKET_LAUNCHER:
            self._secondary_projectile_pool.spawn(
                pos_x=muzzle_x,
                pos_y=muzzle_y,
                angle=angle,
                type_id=1,
            )
        elif player.weapon_id == WeaponId.PULSE_GUN:
            self._secondary_projectile_pool.spawn(
                pos_x=muzzle_x,
                pos_y=muzzle_y,
                angle=angle,
                type_id=4,
            )

        player.ammo = max(0, player.ammo - 1)
        if player.ammo <= 0:
            reload_time = float(weapon.reload_time) if weapon.reload_time is not None else 0.0
            if self._bonus_weapon_power_up_timer > 0.0:
                reload_time *= 0.6
            player.reload_timer_max = max(0.0, reload_time)
            player.reload_timer = player.reload_timer_max
            play_sfx(self._state.audio, resolve_weapon_sfx_ref(weapon.reload_sound), rng=self._state.rng)

    def _update_projectiles(self, dt: float) -> None:
        damage_scale_by_type: dict[int, float] = {}
        for type_id in (0x05, 0x0B, 0x15):
            weapon = weapon_entry_for_projectile_type_id(type_id)
            scale = float(weapon.damage_scale) if weapon is not None and weapon.damage_scale is not None else 0.0
            damage_scale_by_type[type_id] = scale if scale > 0.0 else 1.0

        hits = self._projectile_pool.update(
            dt,
            self._creatures,
            world_size=WORLD_SIZE,
            damage_scale_by_type=damage_scale_by_type,
            rng=self._crand.rand,
        )
        for type_id, origin_x, origin_y, hit_x, hit_y, *_ in hits:
            if type_id == 0x15:
                self._beams.append(
                    DemoBeam(
                        x0=origin_x,
                        y0=origin_y,
                        x1=hit_x,
                        y1=hit_y,
                        life=0.08,
                    )
                )
            if type_id == 0x0B:
                self._explosions.append(
                    DemoExplosion(
                        kind="rocket",
                        x=hit_x,
                        y=hit_y,
                        elapsed=0.0,
                        duration=0.35,
                        max_radius=90.0,
                        damage_per_tick=0.0,
                        tick_interval=1.0,
                    )
                )

        self._secondary_projectile_pool.update_pulse_gun(dt, self._creatures)
        self._creatures = [c for c in self._creatures if c.hp > 0.0]

    def _update_fx(self, dt: float) -> None:
        if self._beams:
            beams: list[DemoBeam] = []
            for beam in self._beams:
                beam.life -= dt
                if beam.life > 0.0:
                    beams.append(beam)
            self._beams = beams

        if not self._explosions:
            return
        survivors: list[DemoExplosion] = []
        for fx in self._explosions:
            fx.elapsed += dt
            if fx.damage_per_tick > 0.0 and fx.tick_interval > 0.0:
                fx.tick_accum += dt
                while fx.tick_accum >= fx.tick_interval:
                    fx.tick_accum -= fx.tick_interval
                    self._apply_explosion_damage(fx)
            if fx.elapsed < fx.duration:
                survivors.append(fx)
        self._explosions = survivors
        self._creatures = [c for c in self._creatures if c.hp > 0.0]

    def _apply_explosion_damage(self, fx: DemoExplosion) -> None:
        t = fx.elapsed / fx.duration if fx.duration > 0 else 1.0
        radius = fx.max_radius * _clamp(t, 0.0, 1.0)
        rsq = radius * radius
        for creature in self._creatures:
            if creature.hp <= 0.0:
                continue
            if _distance_sq(fx.x, fx.y, creature.x, creature.y) <= rsq:
                creature.hp -= fx.damage_per_tick

    def _update_camera(self, dt: float) -> None:
        if not self._players:
            return
        screen_w = float(self._state.config.screen_width)
        screen_h = float(self._state.config.screen_height)
        if screen_w > WORLD_SIZE:
            screen_w = WORLD_SIZE
        if screen_h > WORLD_SIZE:
            screen_h = WORLD_SIZE

        if len(self._players) == 1:
            focus_x = self._players[0].x
            focus_y = self._players[0].y
        else:
            focus_x = sum(p.x for p in self._players) / len(self._players)
            focus_y = sum(p.y for p in self._players) / len(self._players)

        desired_x = (screen_w * 0.5) - focus_x
        desired_y = (screen_h * 0.5) - focus_y

        min_x = screen_w - WORLD_SIZE
        min_y = screen_h - WORLD_SIZE
        if desired_x > -1.0:
            desired_x = -1.0
        if desired_y > -1.0:
            desired_y = -1.0
        if desired_x < min_x:
            desired_x = min_x
        if desired_y < min_y:
            desired_y = min_y

        t = _clamp(dt * 6.0, 0.0, 1.0)
        self._camera_x = _lerp(self._camera_x, desired_x, t)
        self._camera_y = _lerp(self._camera_y, desired_y, t)



from __future__ import annotations

import pyray as rl

from grim.fonts.small import SmallFontData, load_small_font, measure_small_text_width
from grim.view import View, ViewContext

from ..perks import PERK_BY_ID, PerkId, perk_display_description, perk_display_name
from ..ui.perk_menu import (
    PerkMenuAssets,
    PerkMenuLayout,
    UiButtonState,
    button_draw,
    button_update,
    button_width,
    cursor_draw,
    draw_menu_panel,
    draw_menu_item,
    draw_ui_text,
    load_perk_menu_assets,
    menu_item_hit_rect,
    perk_menu_compute_layout,
    ui_origin,
    ui_scale,
    wrap_ui_text,
)
from .registry import register_view


UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_ERROR_COLOR = rl.Color(240, 80, 80, 255)
UI_SPONSOR_COLOR = rl.Color(255, 255, 255, int(255 * 0.5))

PERK_PROMPT_OUTSET_X = 50.0
PERK_PROMPT_BAR_SCALE = 0.75
PERK_PROMPT_BAR_BASE_OFFSET_X = -72.0
PERK_PROMPT_BAR_BASE_OFFSET_Y = -60.0
PERK_PROMPT_BAR_SHIFT_X = -300.0

PERK_PROMPT_LEVEL_UP_SCALE = 0.85
PERK_PROMPT_LEVEL_UP_BASE_OFFSET_X = -230.0
PERK_PROMPT_LEVEL_UP_BASE_OFFSET_Y = -27.0
PERK_PROMPT_LEVEL_UP_BASE_W = 75.0
PERK_PROMPT_LEVEL_UP_BASE_H = 25.0
PERK_PROMPT_LEVEL_UP_SHIFT_X = -46.0
PERK_PROMPT_LEVEL_UP_SHIFT_Y = -4.0

PERK_PROMPT_TEXT_MARGIN_X = 16.0
PERK_PROMPT_TEXT_OFFSET_Y = 8.0


class PerkMenuDebugView:
    def __init__(self, ctx: ViewContext) -> None:
        self._assets_root = ctx.assets_dir
        self._missing_assets: list[str] = []
        self._small: SmallFontData | None = None
        self._assets: PerkMenuAssets | None = None
        self._layout = PerkMenuLayout()

        self._perk_ids = [
            perk_id for perk_id in sorted(PERK_BY_ID.keys()) if perk_id != int(PerkId.ANTIPERK)
        ]
        self._choice_count = 6
        self._selected = 0
        self._expert_owned = False
        self._master_owned = False
        self._show_menu = True
        self._show_prompt = True
        self._panel_slide_x = 0.0
        self._prompt_alpha = 1.0
        self._prompt_pulse = 0.0
        self._prompt_hover = False
        self._prompt_rect: rl.Rectangle | None = None
        self._cancel_button = UiButtonState("Cancel")
        self._debug_overlay = True
        self._show_prompt_rect = False

    def open(self) -> None:
        self._missing_assets.clear()
        try:
            self._small = load_small_font(self._assets_root, self._missing_assets)
        except Exception:
            self._small = None
        self._assets = load_perk_menu_assets(self._assets_root)
        if self._assets.missing:
            self._missing_assets.extend(self._assets.missing)
        rl.hide_cursor()

    def close(self) -> None:
        rl.show_cursor()
        if self._assets is not None:
            self._assets.unload()
            self._assets = None
        if self._small is not None:
            rl.unload_texture(self._small.texture)
            self._small = None

    def _choices(self) -> list[int]:
        if not self._perk_ids:
            return []
        count = max(1, min(self._choice_count, len(self._perk_ids)))
        return self._perk_ids[:count]

    def _prompt_label(self) -> str:
        pending = max(1, int(self._choice_count))
        suffix = f" ({pending})" if pending > 1 else ""
        return f"Press Mouse2 to pick a perk{suffix}"

    @staticmethod
    def _perk_prompt_hinge() -> tuple[float, float]:
        screen_w = float(rl.get_screen_width())
        hinge_x = screen_w + PERK_PROMPT_OUTSET_X
        hinge_y = 80.0 if int(screen_w) == 640 else 40.0
        return hinge_x, hinge_y

    def _perk_prompt_rect(self, label: str) -> rl.Rectangle:
        hinge_x, hinge_y = self._perk_prompt_hinge()
        if self._assets is not None and self._assets.menu_item is not None:
            tex = self._assets.menu_item
            bar_w = float(tex.width) * PERK_PROMPT_BAR_SCALE
            bar_h = float(tex.height) * PERK_PROMPT_BAR_SCALE
            local_x = (PERK_PROMPT_BAR_BASE_OFFSET_X + PERK_PROMPT_BAR_SHIFT_X) * PERK_PROMPT_BAR_SCALE
            local_y = PERK_PROMPT_BAR_BASE_OFFSET_Y * PERK_PROMPT_BAR_SCALE
            return rl.Rectangle(
                float(hinge_x + local_x),
                float(hinge_y + local_y),
                float(bar_w),
                float(bar_h),
            )

        text_w = float(_ui_text_width(self._small, label, 1.0))
        text_h = 20.0
        x = float(rl.get_screen_width()) - PERK_PROMPT_TEXT_MARGIN_X - text_w
        y = hinge_y + PERK_PROMPT_TEXT_OFFSET_Y
        return rl.Rectangle(x, y, text_w, text_h)

    def _draw_perk_prompt(self) -> None:
        if not self._show_prompt:
            return
        if self._assets is None:
            return
        label = self._prompt_label()
        if not label:
            return
        alpha = max(0.0, min(self._prompt_alpha, 1.0))
        if alpha <= 1e-3:
            return

        hinge_x, hinge_y = self._perk_prompt_hinge()
        rot_deg = (1.0 - alpha) * 90.0
        tint = rl.Color(255, 255, 255, int(255 * alpha))

        text_w = float(_ui_text_width(self._small, label, 1.0))
        x = float(rl.get_screen_width()) - PERK_PROMPT_TEXT_MARGIN_X - text_w
        y = hinge_y + PERK_PROMPT_TEXT_OFFSET_Y
        color = rl.Color(UI_TEXT_COLOR.r, UI_TEXT_COLOR.g, UI_TEXT_COLOR.b, int(255 * alpha))
        draw_ui_text(self._small, label, x, y, scale=1.0, color=color)

        if self._assets.menu_item is not None:
            tex = self._assets.menu_item
            bar_w = float(tex.width) * PERK_PROMPT_BAR_SCALE
            bar_h = float(tex.height) * PERK_PROMPT_BAR_SCALE
            local_x = (PERK_PROMPT_BAR_BASE_OFFSET_X + PERK_PROMPT_BAR_SHIFT_X) * PERK_PROMPT_BAR_SCALE
            local_y = PERK_PROMPT_BAR_BASE_OFFSET_Y * PERK_PROMPT_BAR_SCALE
            src = rl.Rectangle(float(tex.width), 0.0, -float(tex.width), float(tex.height))
            dst = rl.Rectangle(float(hinge_x), float(hinge_y), float(bar_w), float(bar_h))
            origin = rl.Vector2(float(-local_x), float(-local_y))
            rl.draw_texture_pro(tex, src, dst, origin, rot_deg, tint)

        if self._assets.title_level_up is not None:
            tex = self._assets.title_level_up
            local_x = PERK_PROMPT_LEVEL_UP_BASE_OFFSET_X * PERK_PROMPT_LEVEL_UP_SCALE + PERK_PROMPT_LEVEL_UP_SHIFT_X
            local_y = PERK_PROMPT_LEVEL_UP_BASE_OFFSET_Y * PERK_PROMPT_LEVEL_UP_SCALE + PERK_PROMPT_LEVEL_UP_SHIFT_Y
            w = PERK_PROMPT_LEVEL_UP_BASE_W * PERK_PROMPT_LEVEL_UP_SCALE
            h = PERK_PROMPT_LEVEL_UP_BASE_H * PERK_PROMPT_LEVEL_UP_SCALE
            pulse_alpha = (100.0 + float(int(self._prompt_pulse * 155.0 / 1000.0))) / 255.0
            pulse_alpha = max(0.0, min(1.0, pulse_alpha))
            label_alpha = max(0.0, min(1.0, alpha * pulse_alpha))
            pulse_tint = rl.Color(255, 255, 255, int(255 * label_alpha))
            src = rl.Rectangle(0.0, 0.0, float(tex.width), float(tex.height))
            dst = rl.Rectangle(float(hinge_x), float(hinge_y), float(w), float(h))
            origin = rl.Vector2(float(-local_x), float(-local_y))
            rl.draw_texture_pro(tex, src, dst, origin, rot_deg, pulse_tint)
            if label_alpha > 0.0:
                rl.begin_blend_mode(rl.BLEND_ADDITIVE)
                rl.draw_texture_pro(tex, src, dst, origin, rot_deg, pulse_tint)
                rl.end_blend_mode()

    def update(self, dt: float) -> None:
        dt_ms = float(min(dt, 0.1) * 1000.0)

        if rl.is_key_pressed(rl.KeyboardKey.KEY_F1):
            self._debug_overlay = not self._debug_overlay
        if rl.is_key_pressed(rl.KeyboardKey.KEY_E):
            self._expert_owned = not self._expert_owned
            if not self._expert_owned:
                self._master_owned = False
        if rl.is_key_pressed(rl.KeyboardKey.KEY_M):
            self._master_owned = not self._master_owned
            if self._master_owned:
                self._expert_owned = True
        if rl.is_key_pressed(rl.KeyboardKey.KEY_O):
            self._show_menu = not self._show_menu
        if rl.is_key_pressed(rl.KeyboardKey.KEY_P):
            self._show_prompt = not self._show_prompt
        if rl.is_key_pressed(rl.KeyboardKey.KEY_H):
            self._show_prompt_rect = not self._show_prompt_rect
        if rl.is_key_pressed(rl.KeyboardKey.KEY_R):
            self._panel_slide_x = 0.0
            self._choice_count = 6
            self._selected = 0
            self._expert_owned = False
            self._master_owned = False
            self._show_menu = True
            self._show_prompt = True
            self._prompt_pulse = 0.0
            self._prompt_hover = False

        if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT_BRACKET):
            self._choice_count = max(1, self._choice_count - 1)
            self._selected = min(self._selected, self._choice_count - 1)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT_BRACKET):
            self._choice_count = min(len(self._perk_ids), self._choice_count + 1)

        if self._show_menu and self._choice_count > 0:
            if rl.is_key_pressed(rl.KeyboardKey.KEY_DOWN):
                self._selected = (self._selected + 1) % self._choice_count
            if rl.is_key_pressed(rl.KeyboardKey.KEY_UP):
                self._selected = (self._selected - 1) % self._choice_count

        step = 10.0
        if rl.is_key_down(rl.KeyboardKey.KEY_LEFT_SHIFT) or rl.is_key_down(rl.KeyboardKey.KEY_RIGHT_SHIFT):
            step = 40.0
        if rl.is_key_down(rl.KeyboardKey.KEY_LEFT):
            self._panel_slide_x -= step
        if rl.is_key_down(rl.KeyboardKey.KEY_RIGHT):
            self._panel_slide_x += step

        self._panel_slide_x = _clamp(self._panel_slide_x, -self._layout.panel_w, 0.0)

        self._prompt_hover = False
        self._prompt_rect = None
        if self._show_prompt:
            label = self._prompt_label()
            if label:
                rect = self._perk_prompt_rect(label)
                self._prompt_rect = rect
                mouse = rl.get_mouse_position()
                self._prompt_hover = rl.check_collision_point_rec(mouse, rect)

        pulse_delta = dt_ms * (6.0 if self._prompt_hover else -2.0)
        self._prompt_pulse = _clamp(self._prompt_pulse + pulse_delta, 0.0, 1000.0)

        if not self._show_menu or self._assets is None:
            return

        choices = self._choices()
        if not choices:
            return
        screen_w = float(rl.get_screen_width())
        screen_h = float(rl.get_screen_height())
        scale = ui_scale(screen_w, screen_h)
        origin_x, origin_y = ui_origin(screen_w, screen_h, scale)
        computed = perk_menu_compute_layout(
            self._layout,
            screen_w=screen_w,
            origin_x=origin_x,
            origin_y=origin_y,
            scale=scale,
            choice_count=len(choices),
            expert_owned=self._expert_owned,
            master_owned=self._master_owned,
            panel_slide_x=self._panel_slide_x,
        )

        mouse = rl.get_mouse_position()
        click = rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
        for idx, perk_id in enumerate(choices):
            label = perk_display_name(int(perk_id))
            item_x = computed.list_x
            item_y = computed.list_y + float(idx) * computed.list_step_y
            rect = menu_item_hit_rect(self._small, label, x=item_x, y=item_y, scale=scale)
            if rl.check_collision_point_rec(mouse, rect):
                self._selected = idx
                break

        cancel_w = button_width(self._small, self._cancel_button.label, scale=scale, force_wide=self._cancel_button.force_wide)
        cancel_x = computed.cancel_x
        button_y = computed.cancel_y
        if button_update(
            self._cancel_button,
            x=cancel_x,
            y=button_y,
            width=cancel_w,
            dt_ms=dt_ms,
            mouse=mouse,
            click=click,
        ):
            self._show_menu = False

    def draw(self) -> None:
        rl.clear_background(rl.Color(0, 0, 0, 255))
        if self._missing_assets and self._debug_overlay:
            draw_ui_text(
                self._small,
                "Missing assets: " + ", ".join(self._missing_assets),
                24.0,
                24.0,
                scale=1.0,
                color=UI_ERROR_COLOR,
            )
            return

        self._draw_perk_prompt()
        if self._show_prompt_rect and self._prompt_rect is not None:
            rl.draw_rectangle_lines_ex(self._prompt_rect, 1.0, rl.Color(255, 0, 255, 255))

        if self._show_menu and self._assets is not None:
            choices = self._choices()
            if choices:
                screen_w = float(rl.get_screen_width())
                screen_h = float(rl.get_screen_height())
                scale = ui_scale(screen_w, screen_h)
                origin_x, origin_y = ui_origin(screen_w, screen_h, scale)
                computed = perk_menu_compute_layout(
                    self._layout,
                    screen_w=screen_w,
                    origin_x=origin_x,
                    origin_y=origin_y,
                    scale=scale,
                    choice_count=len(choices),
                    expert_owned=self._expert_owned,
                    master_owned=self._master_owned,
                    panel_slide_x=self._panel_slide_x,
                )

                if self._assets.menu_panel is not None:
                    draw_menu_panel(self._assets.menu_panel, dst=computed.panel)

                if self._assets.title_pick_perk is not None:
                    tex = self._assets.title_pick_perk
                    src = rl.Rectangle(0.0, 0.0, float(tex.width), float(tex.height))
                    rl.draw_texture_pro(tex, src, computed.title, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)

                sponsor = None
                if self._master_owned:
                    sponsor = "extra perks sponsored by the Perk Master"
                elif self._expert_owned:
                    sponsor = "extra perk sponsored by the Perk Expert"
                if sponsor:
                    draw_ui_text(
                        self._small,
                        sponsor,
                        computed.sponsor_x,
                        computed.sponsor_y,
                        scale=scale,
                        color=UI_SPONSOR_COLOR,
                    )

                mouse = rl.get_mouse_position()
                for idx, perk_id in enumerate(choices):
                    label = perk_display_name(int(perk_id))
                    item_x = computed.list_x
                    item_y = computed.list_y + float(idx) * computed.list_step_y
                    rect = menu_item_hit_rect(self._small, label, x=item_x, y=item_y, scale=scale)
                    hovered = rl.check_collision_point_rec(mouse, rect) or (idx == self._selected)
                    draw_menu_item(self._small, label, x=item_x, y=item_y, scale=scale, hovered=hovered)

                selected_id = choices[self._selected]
                desc = perk_display_description(int(selected_id))
                desc_x = float(computed.desc.x)
                desc_y = float(computed.desc.y)
                desc_w = float(computed.desc.width)
                desc_h = float(computed.desc.height)
                desc_scale = scale * 0.85
                desc_lines = wrap_ui_text(self._small, desc, max_width=desc_w, scale=desc_scale)
                line_h = float(self._small.cell_size * desc_scale) if self._small is not None else float(20 * desc_scale)
                y = desc_y
                for line in desc_lines:
                    if y + line_h > desc_y + desc_h:
                        break
                    draw_ui_text(self._small, line, desc_x, y, scale=desc_scale, color=UI_TEXT_COLOR)
                    y += line_h

                cancel_w = button_width(self._small, self._cancel_button.label, scale=scale, force_wide=self._cancel_button.force_wide)
                button_draw(self._assets, self._small, self._cancel_button, x=computed.cancel_x, y=computed.cancel_y, width=cancel_w, scale=scale)

        screen_w = float(rl.get_screen_width())
        screen_h = float(rl.get_screen_height())
        scale = ui_scale(screen_w, screen_h)
        if self._assets is not None:
            cursor_draw(self._assets, mouse=rl.get_mouse_position(), scale=scale)

        if self._debug_overlay:
            self._draw_overlay()

    def _draw_overlay(self) -> None:
        x = 24.0
        y = 24.0
        scale = 0.9
        line_h = float(self._small.cell_size * scale) if self._small is not None else float(20 * scale)
        lines = [
            "Perk menu render debug (F1 hide)",
            "O toggle menu  P toggle prompt  H hover rect  E/M toggle Expert/Master",
            "Left/Right slide_x (hold Shift for bigger)  [/] choices  Up/Down selection  R reset",
            f"slide_x={self._panel_slide_x:.1f} choices={self._choice_count} selected={self._selected}",
        ]
        for line in lines:
            draw_ui_text(self._small, line, x, y, scale=scale, color=UI_HINT_COLOR)
            y += line_h


def _clamp(value: float, lo: float, hi: float) -> float:
    if value < lo:
        return lo
    if value > hi:
        return hi
    return value


def _ui_text_width(font: SmallFontData | None, text: str, scale: float) -> float:
    if font is None:
        return float(rl.measure_text(text, int(20 * scale)))
    return float(measure_small_text_width(font, text, scale))


@register_view("perk-menu-debug", "Perk menu render debug")
def _create_perk_menu_debug_view(*, ctx: ViewContext) -> View:
    return PerkMenuDebugView(ctx)



from __future__ import annotations

import math
from typing import Callable

from ..gameplay import PlayerState, perk_active
from ..perks import PerkId
from .runtime import CREATURE_HITBOX_ALIVE, CreatureState
from .spawn import CreatureFlags


def _owner_id_to_player_index(owner_id: int) -> int | None:
    if owner_id == -100:
        return 0
    if owner_id < 0:
        return -1 - owner_id
    return None


def creature_apply_damage(
    creature: CreatureState,
    *,
    damage_amount: float,
    damage_type: int,
    impulse_x: float,
    impulse_y: float,
    owner_id: int,
    dt: float,
    players: list[PlayerState],
    rand: Callable[[], int],
) -> bool:
    """Apply damage to a creature, returning True if the hit killed it.

    This is a partial port of `creature_apply_damage` (FUN_004207c0).

    Notes:
    - Death side-effects are handled by the caller (see Phase 2 in `plan.md`).
    - `damage_type` is a native integer category; call sites must supply it.
    """

    creature.last_hit_owner_id = int(owner_id)
    creature.hit_flash_timer = 0.2

    player_index = _owner_id_to_player_index(owner_id)
    attacker = players[player_index] if player_index is not None and 0 <= player_index < len(players) else None

    damage = float(damage_amount)

    if int(damage_type) == 1 and attacker is not None:
        if perk_active(attacker, PerkId.URANIUM_FILLED_BULLETS):
            damage *= 2.0

        if perk_active(attacker, PerkId.LIVING_FORTRESS):
            for player in players:
                if float(player.health) <= 0.0:
                    continue
                timer = float(player.living_fortress_timer)
                if timer > 0.0:
                    damage *= timer * 0.05 + 1.0

        if perk_active(attacker, PerkId.BARREL_GREASER):
            damage *= 1.4
        if perk_active(attacker, PerkId.DOCTOR):
            damage *= 1.2

        if (creature.flags & CreatureFlags.ANIM_PING_PONG) == 0:
            jitter = float((int(rand()) & 0x7F) - 0x40) * 0.002
            size = max(1e-6, float(creature.size))
            turn = jitter / (size * 0.025)
            turn = max(-math.pi / 2.0, min(math.pi / 2.0, turn))
            creature.heading += turn

    if int(damage_type) == 7 and attacker is not None:
        if perk_active(attacker, PerkId.ION_GUN_MASTER):
            damage *= 1.2

    if creature.hp <= 0.0:
        if dt > 0.0:
            creature.hitbox_size -= float(dt) * 15.0
        return True

    if int(damage_type) == 4 and attacker is not None:
        if perk_active(attacker, PerkId.PYROMANIAC):
            damage *= 1.5
            rand()

    creature.hp -= damage
    creature.vel_x -= float(impulse_x)
    creature.vel_y -= float(impulse_y)

    if creature.hp <= 0.0:
        if dt > 0.0:
            creature.hitbox_size = float(creature.hitbox_size) - float(dt)
        else:
            creature.hitbox_size = float(creature.hitbox_size) - 0.001
        creature.vel_x -= float(impulse_x) * 2.0
        creature.vel_y -= float(impulse_y) * 2.0
        return True

    if creature.hitbox_size != CREATURE_HITBOX_ALIVE and dt > 0.0:
        creature.hitbox_size = CREATURE_HITBOX_ALIVE

    return False



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import random
from typing import Iterable

import pyray as rl

from .console import ConsoleState
from . import paq
from . import sfx_map


SFX_PAK_NAME = "sfx.paq"
DEFAULT_VOICE_COUNT = 4


@dataclass(slots=True)
class SfxSample:
    entry_name: str
    source: rl.Sound
    aliases: list[rl.Sound]
    next_voice: int = 0

    def voices(self) -> Iterable[rl.Sound]:
        yield self.source
        yield from self.aliases

    def acquire_voice(self) -> rl.Sound:
        for voice in self.voices():
            if not rl.is_sound_playing(voice):
                return voice
        voices = [self.source, *self.aliases]
        idx = self.next_voice % len(voices)
        self.next_voice += 1
        return voices[idx]


@dataclass(slots=True)
class SfxState:
    ready: bool
    enabled: bool
    volume: float
    voice_count: int
    entries: dict[str, bytes]
    directory: Path | None
    key_to_entry: dict[str, str]
    variants: dict[str, tuple[str, ...]]
    samples: dict[str, SfxSample]
    missing_keys: set[str]


def init_sfx_state(
    *,
    ready: bool,
    enabled: bool,
    volume: float,
    voice_count: int = DEFAULT_VOICE_COUNT,
) -> SfxState:
    return SfxState(
        ready=ready,
        enabled=enabled,
        volume=float(volume),
        voice_count=max(1, int(voice_count)),
        entries={},
        directory=None,
        key_to_entry={},
        variants={},
        samples={},
        missing_keys=set(),
    )


def _derive_sfx_key(entry_name: str) -> str:
    return "sfx_" + Path(entry_name).stem.lower()


def _derive_sfx_base(key: str) -> str | None:
    if not key.startswith("sfx_"):
        return None
    stem = key[4:]
    if len(stem) < 3:
        return None
    if "_" not in stem:
        return None
    base, suffix = stem.rsplit("_", 1)
    if not suffix.isdigit():
        return None
    return "sfx_" + base


def _build_variants(keys: Iterable[str]) -> dict[str, tuple[str, ...]]:
    base_to_keys: dict[str, list[str]] = {}
    for key in keys:
        base = _derive_sfx_base(key)
        if base is None:
            continue
        base_to_keys.setdefault(base, []).append(key)
    return {base: tuple(sorted(values)) for base, values in base_to_keys.items()}


def load_sfx_index(state: SfxState, assets_dir: Path, console: ConsoleState) -> None:
    if not state.ready or not state.enabled:
        return

    sfx_dir = assets_dir / "sfx"
    if sfx_dir.exists() and sfx_dir.is_dir():
        entry_names: list[str] = []
        for path in sorted(sfx_dir.iterdir()):
            if not path.is_file():
                continue
            if path.suffix.lower() not in {".ogg", ".wav"}:
                continue
            entry_names.append(path.name)

        state.directory = sfx_dir
        state.entries.clear()
        available = set(entry_names)
        state.key_to_entry = {_derive_sfx_key(name): name for name in entry_names}
        for key, name in sfx_map.SFX_ENTRY_BY_KEY.items():
            if name in available:
                state.key_to_entry[key] = name
        state.variants = _build_variants(state.key_to_entry.keys())
        console.log.log(f"audio: sfx indexed {len(entry_names)} files from {sfx_dir}")
        console.log.flush()
        return

    paq_path = assets_dir / SFX_PAK_NAME
    if not paq_path.exists():
        raise FileNotFoundError(f"audio: missing {SFX_PAK_NAME} in {assets_dir}")
    entries: dict[str, bytes] = {}
    for name, data in paq.iter_entries(paq_path):
        entries[name.replace("\\", "/")] = data
    state.directory = None
    state.entries = entries
    available = set(entries.keys())
    state.key_to_entry = {_derive_sfx_key(name): name for name in entries.keys()}
    for key, name in sfx_map.SFX_ENTRY_BY_KEY.items():
        if name in available:
            state.key_to_entry[key] = name
    state.variants = _build_variants(state.key_to_entry.keys())
    console.log.log(f"audio: sfx indexed {len(entries)} entries from {SFX_PAK_NAME}")
    console.log.flush()


def _normalize_sfx_key(state: SfxState, key: str) -> str | None:
    key = key.strip().lstrip("_")
    if not key:
        return None
    key = sfx_map.SFX_KEY_ALIASES.get(key, key)
    if "_alias_" in key:
        key = key.split("_alias_", 1)[0]
    if key in state.key_to_entry:
        return key
    if key.endswith("_alt"):
        cand = key[: -len("_alt")]
        if cand in state.key_to_entry:
            return cand
    return None


def _load_sample(state: SfxState, key: str) -> SfxSample | None:
    resolved = _normalize_sfx_key(state, key)
    if resolved is None:
        return None
    existing = state.samples.get(resolved)
    if existing is not None:
        return existing

    entry_name = state.key_to_entry.get(resolved)
    if entry_name is None:
        return None

    if state.directory is not None:
        path = state.directory / entry_name
        source = rl.load_sound(str(path))
    else:
        data = state.entries.get(entry_name)
        if data is None:
            return None
        file_type = Path(entry_name).suffix.lower()
        wave = rl.load_wave_from_memory(file_type, data, len(data))
        source = rl.load_sound_from_wave(wave)
        rl.unload_wave(wave)

    aliases = [rl.load_sound_alias(source) for _ in range(max(1, state.voice_count) - 1)]

    sample = SfxSample(entry_name=entry_name, source=source, aliases=aliases)
    for voice in sample.voices():
        rl.set_sound_volume(voice, state.volume)
    state.samples[resolved] = sample
    return sample


def play_sfx(
    state: SfxState | None,
    key: str | None,
    *,
    rng: random.Random | None = None,
    allow_variants: bool = True,
) -> None:
    if state is None or not state.ready or not state.enabled:
        return
    if not key:
        return

    resolved = _normalize_sfx_key(state, key)
    if resolved is None:
        state.missing_keys.add(key)
        return

    if allow_variants:
        base = _derive_sfx_base(resolved) or resolved
        variants = state.variants.get(base)
        if variants:
            rng = rng or random
            resolved = rng.choice(variants)

    sample = _load_sample(state, resolved)
    if sample is None:
        state.missing_keys.add(resolved)
        return
    rl.play_sound(sample.acquire_voice())


def sfx_key_for_id(sfx_id: int) -> str | None:
    if sfx_id < 0:
        return None
    if sfx_id >= len(sfx_map.SFX_KEY_BY_ID):
        return None
    return sfx_map.SFX_KEY_BY_ID[sfx_id]


def play_sfx_id(state: SfxState | None, sfx_id: int, *, rng: random.Random | None = None) -> None:
    key = sfx_key_for_id(int(sfx_id))
    if key is None:
        return
    play_sfx(state, key, rng=rng, allow_variants=False)


def set_sfx_volume(state: SfxState | None, volume: float) -> None:
    if state is None:
        return
    volume = float(volume)
    if volume < 0.0:
        volume = 0.0
    if volume > 1.0:
        volume = 1.0
    state.volume = volume
    for sample in state.samples.values():
        for voice in sample.voices():
            rl.set_sound_volume(voice, state.volume)


def shutdown_sfx(state: SfxState) -> None:
    if not state.ready:
        return
    for sample in state.samples.values():
        for alias in sample.aliases:
            try:
                rl.stop_sound(alias)
                rl.unload_sound_alias(alias)
            except Exception:
                pass
        try:
            rl.stop_sound(sample.source)
            rl.unload_sound(sample.source)
        except Exception:
            pass
    state.samples.clear()
    state.entries.clear()
    state.key_to_entry.clear()
    state.variants.clear()
    state.missing_keys.clear()
    state.directory = None



from __future__ import annotations

from dataclasses import dataclass

__all__ = [
    "EFFECT_ID_ATLAS_TABLE",
    "EFFECT_ID_ATLAS_TABLE_BY_ID",
    "EffectAtlasEntry",
    "SIZE_CODE_GRID",
    "effect_src_rect",
]


SIZE_CODE_GRID: dict[int, int] = {
    0x10: 16,
    0x20: 8,
    0x40: 4,
    0x80: 2,
}


@dataclass(frozen=True, slots=True)
class EffectAtlasEntry:
    effect_id: int
    size_code: int
    frame: int

    @property
    def grid(self) -> int:
        return SIZE_CODE_GRID[self.size_code]


# Extracted from `effect_id_table` (`size_code`, `frame`) (see `docs/structs/effects.md`).
EFFECT_ID_ATLAS_TABLE: tuple[EffectAtlasEntry, ...] = (
    EffectAtlasEntry(0x00, 0x80, 0x02),
    EffectAtlasEntry(0x01, 0x80, 0x03),
    EffectAtlasEntry(0x02, 0x20, 0x00),
    EffectAtlasEntry(0x03, 0x20, 0x01),
    EffectAtlasEntry(0x04, 0x20, 0x02),
    EffectAtlasEntry(0x05, 0x20, 0x03),
    EffectAtlasEntry(0x06, 0x20, 0x04),
    EffectAtlasEntry(0x07, 0x20, 0x05),
    EffectAtlasEntry(0x08, 0x20, 0x08),
    EffectAtlasEntry(0x09, 0x20, 0x09),
    EffectAtlasEntry(0x0A, 0x20, 0x0A),
    EffectAtlasEntry(0x0B, 0x20, 0x0B),
    EffectAtlasEntry(0x0C, 0x40, 0x05),
    EffectAtlasEntry(0x0D, 0x40, 0x03),
    EffectAtlasEntry(0x0E, 0x40, 0x04),
    EffectAtlasEntry(0x0F, 0x40, 0x05),
    EffectAtlasEntry(0x10, 0x40, 0x06),
    EffectAtlasEntry(0x11, 0x40, 0x07),
    EffectAtlasEntry(0x12, 0x10, 0x26),
)

EFFECT_ID_ATLAS_TABLE_BY_ID: dict[int, EffectAtlasEntry] = {entry.effect_id: entry for entry in EFFECT_ID_ATLAS_TABLE}


def effect_src_rect(effect_id: int, *, texture_width: float, texture_height: float) -> tuple[float, float, float, float] | None:
    entry = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(effect_id))
    if entry is None:
        return None

    grid = SIZE_CODE_GRID.get(entry.size_code)
    if not grid:
        return None

    frame = int(entry.frame)
    col = frame % grid
    row = frame // grid
    cell_w = float(texture_width) / float(grid)
    cell_h = float(texture_height) / float(grid)
    return cell_w * float(col), cell_h * float(row), cell_w, cell_h



from __future__ import annotations

from dataclasses import dataclass
import math
import random
from typing import Iterator

from ..creatures.spawn import SpawnId
from .types import SpawnEntry


@dataclass(frozen=True, slots=True)
class EdgePoints:
    left: tuple[float, float]
    right: tuple[float, float]
    top: tuple[float, float]
    bottom: tuple[float, float]


def center_point(width: float, height: float | None = None) -> tuple[float, float]:
    if height is None:
        height = width
    return float(width) / 2.0, float(height) / 2.0


def edge_midpoints(width: float, height: float | None = None, offset: float = 64.0) -> EdgePoints:
    if height is None:
        height = width
    cx, cy = center_point(width, height)
    return EdgePoints(
        left=(-offset, cy),
        right=(float(width) + offset, cy),
        top=(cx, -offset),
        bottom=(cx, float(height) + offset),
    )


def corner_points(width: float, height: float | None = None, offset: float = 64.0) -> tuple[tuple[float, float], ...]:
    if height is None:
        height = width
    return (
        (-offset, -offset),
        (float(width) + offset, -offset),
        (-offset, float(height) + offset),
        (float(width) + offset, float(height) + offset),
    )


def iter_angles(count: int, *, step: float | None = None, start: float = 0.0) -> Iterator[float]:
    if count <= 0:
        return iter(())
    if step is None:
        step = math.tau / float(count)
    for idx in range(count):
        yield start + float(idx) * step


def ring_points(
    center_x: float,
    center_y: float,
    radius: float,
    count: int,
    *,
    step: float | None = None,
    start: float = 0.0,
) -> Iterator[tuple[float, float, float]]:
    for angle in iter_angles(count, step=step, start=start):
        yield (
            math.cos(angle) * radius + center_x,
            math.sin(angle) * radius + center_y,
            angle,
        )


def random_angle(rng: random.Random) -> float:
    return float(rng.randrange(0x264)) * 0.01


def radial_points(
    center_x: float,
    center_y: float,
    angle: float,
    radius_start: float,
    radius_end: float,
    radius_step: float,
) -> Iterator[tuple[float, float]]:
    cos_a = math.cos(angle)
    sin_a = math.sin(angle)
    radius = radius_start
    while radius < radius_end:
        yield (
            cos_a * radius + center_x,
            sin_a * radius + center_y,
        )
        radius += radius_step


def heading_from_center(x: float, y: float, center_x: float, center_y: float) -> float:
    return math.atan2(y - center_y, x - center_x) - (math.pi / 2.0)


def line_points_x(start: float, step: float, count: int, y: float) -> Iterator[tuple[float, float]]:
    for idx in range(count):
        yield start + float(idx) * step, y


def line_points_y(start: float, step: float, count: int, x: float) -> Iterator[tuple[float, float]]:
    for idx in range(count):
        yield x, start + float(idx) * step


def spawn(
    *,
    x: float,
    y: float,
    heading: float = 0.0,
    spawn_id: SpawnId,
    trigger_ms: int,
    count: int,
) -> SpawnEntry:
    return SpawnEntry(
        x=x,
        y=y,
        heading=heading,
        spawn_id=spawn_id,
        trigger_ms=trigger_ms,
        count=count,
    )


def spawn_at(
    point: tuple[float, float],
    *,
    heading: float = 0.0,
    spawn_id: SpawnId,
    trigger_ms: int,
    count: int,
) -> SpawnEntry:
    x, y = point
    return spawn(
        x=x,
        y=y,
        heading=heading,
        spawn_id=spawn_id,
        trigger_ms=trigger_ms,
        count=count,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import pyray as rl

from grim.audio import set_music_volume, set_sfx_volume
from grim.config import apply_detail_preset
from grim.fonts.small import SmallFontData, draw_small_text, load_small_font, measure_small_text_width

from ..menu import (
    MENU_LABEL_ROW_HEIGHT,
    MENU_LABEL_ROW_OPTIONS,
    MENU_LABEL_WIDTH,
    MENU_PANEL_OFFSET_X,
    MENU_PANEL_OFFSET_Y,
    MENU_PANEL_WIDTH,
    MenuView,
    _draw_menu_cursor,
)
from ..transitions import _draw_screen_fade
from .base import PANEL_TIMELINE_END_MS, PANEL_TIMELINE_START_MS, PanelMenuView

if TYPE_CHECKING:
    from ...game import GameState


@dataclass(slots=True)
class SliderState:
    value: int
    min_value: int
    max_value: int


class OptionsMenuView(PanelMenuView):
    _LABELS = (
        "Sound volume:",
        "Music volume:",
        "Graphics detail:",
        "Mouse sensitivity:",
    )

    def __init__(self, state: GameState) -> None:
        super().__init__(state, title="Options")
        self._small_font: SmallFontData | None = None
        self._rect_on: rl.Texture2D | None = None
        self._rect_off: rl.Texture2D | None = None
        self._check_on: rl.Texture2D | None = None
        self._check_off: rl.Texture2D | None = None
        self._button_tex: rl.Texture2D | None = None
        self._slider_sfx = SliderState(10, 0, 10)
        self._slider_music = SliderState(10, 0, 10)
        self._slider_detail = SliderState(5, 1, 5)
        self._slider_mouse = SliderState(10, 1, 10)
        self._ui_info_texts = True
        self._active_slider: str | None = None
        self._dirty = False

    def open(self) -> None:
        super().open()
        cache = self._ensure_cache()
        self._rect_on = cache.get_or_load("ui_rectOn", "ui/ui_rectOn.jaz").texture
        self._rect_off = cache.get_or_load("ui_rectOff", "ui/ui_rectOff.jaz").texture
        self._check_on = cache.get_or_load("ui_checkOn", "ui/ui_checkOn.jaz").texture
        self._check_off = cache.get_or_load("ui_checkOff", "ui/ui_checkOff.jaz").texture
        self._button_tex = cache.get_or_load("ui_button_md", "ui/ui_button_145x32.jaz").texture
        self._active_slider = None
        self._dirty = False
        self._sync_from_config()

    def update(self, dt: float) -> None:
        super().update(dt)
        if self._closing:
            return
        entry = self._entry
        if entry is None or not self._entry_enabled(entry):
            return

        config = self._state.config
        layout = self._content_layout()
        label_x = layout["label_x"]
        base_y = layout["base_y"]
        scale = layout["scale"]
        slider_x = layout["slider_x"]

        rect_on = self._rect_on
        rect_off = self._rect_off
        if rect_on is None or rect_off is None:
            return

        if self._update_slider("sfx", self._slider_sfx, slider_x, base_y + 47.0 * scale, rect_on, rect_off, scale):
            config.data["sfx_volume"] = float(self._slider_sfx.value) * 0.1
            set_sfx_volume(self._state.audio, float(config.data["sfx_volume"]))
            self._dirty = True

        if self._update_slider("music", self._slider_music, slider_x, base_y + 67.0 * scale, rect_on, rect_off, scale):
            config.data["music_volume"] = float(self._slider_music.value) * 0.1
            set_music_volume(self._state.audio, float(config.data["music_volume"]))
            self._dirty = True

        if self._update_slider("detail", self._slider_detail, slider_x, base_y + 87.0 * scale, rect_on, rect_off, scale):
            preset = apply_detail_preset(config, self._slider_detail.value)
            self._slider_detail.value = preset
            self._dirty = True

        if self._update_slider("mouse", self._slider_mouse, slider_x, base_y + 107.0 * scale, rect_on, rect_off, scale):
            sensitivity = float(self._slider_mouse.value) * 0.1
            if sensitivity < 0.1:
                sensitivity = 0.1
            if sensitivity > 1.0:
                sensitivity = 1.0
            config.data["mouse_sensitivity"] = sensitivity
            self._dirty = True

        if self._update_checkbox(label_x, base_y + 135.0 * scale, scale):
            value = 1 if self._ui_info_texts else 0
            config.data["ui_info_texts"] = value
            self._dirty = True

        if self._update_controls_button(label_x - 8.0 * scale, base_y + 155.0 * scale, scale):
            self._begin_close_transition("open_controls")

    def draw(self) -> None:
        rl.clear_background(rl.BLACK)
        if self._ground is not None:
            self._ground.draw(0.0, 0.0)
        _draw_screen_fade(self._state)
        assets = self._assets
        entry = self._entry
        if assets is None or entry is None:
            return
        self._draw_panel()
        self._draw_entry(entry)
        self._draw_sign()
        self._draw_options_contents()
        _draw_menu_cursor(self._state, pulse_time=self._cursor_pulse_time)

    def _begin_close_transition(self, action: str) -> None:
        if self._dirty:
            try:
                self._state.config.save()
            except Exception:
                pass
            self._dirty = False
        super()._begin_close_transition(action)

    def _ensure_small_font(self) -> SmallFontData:
        if self._small_font is not None:
            return self._small_font
        missing_assets: list[str] = []
        self._small_font = load_small_font(self._state.assets_dir, missing_assets)
        return self._small_font

    def _sync_from_config(self) -> None:
        config = self._state.config
        self._ui_info_texts = bool(int(config.data.get("ui_info_texts", 1) or 0))

        sfx_volume = float(config.data.get("sfx_volume", 1.0))
        music_volume = float(config.data.get("music_volume", 1.0))
        detail_preset = int(config.data.get("detail_preset", 5))
        mouse_sensitivity = float(config.data.get("mouse_sensitivity", 1.0))

        self._slider_sfx.value = max(self._slider_sfx.min_value, min(self._slider_sfx.max_value, int(sfx_volume * 10.0)))
        self._slider_music.value = max(
            self._slider_music.min_value, min(self._slider_music.max_value, int(music_volume * 10.0))
        )
        if detail_preset < self._slider_detail.min_value:
            detail_preset = self._slider_detail.min_value
        if detail_preset > self._slider_detail.max_value:
            detail_preset = self._slider_detail.max_value
        self._slider_detail.value = detail_preset
        self._slider_mouse.value = max(
            self._slider_mouse.min_value,
            min(self._slider_mouse.max_value, int(mouse_sensitivity * 10.0 + 0.5)),
        )

    def _content_layout(self) -> dict[str, float]:
        panel_scale, _local_shift = self._menu_item_scale(0)
        panel_w = MENU_PANEL_WIDTH * panel_scale
        _angle_rad, slide_x = MenuView._ui_element_anim(
            self,
            index=1,
            start_ms=PANEL_TIMELINE_START_MS,
            end_ms=PANEL_TIMELINE_END_MS,
            width=panel_w,
        )
        panel_x = self._panel_pos_x + slide_x
        panel_y = self._panel_pos_y + self._widescreen_y_shift
        origin_x = -(MENU_PANEL_OFFSET_X * panel_scale)
        origin_y = -(MENU_PANEL_OFFSET_Y * panel_scale)
        panel_left = panel_x - origin_x
        panel_top = panel_y - origin_y
        base_x = panel_left + 212.0 * panel_scale
        base_y = panel_top + 32.0 * panel_scale
        label_x = base_x + 8.0 * panel_scale
        slider_x = label_x + 130.0 * panel_scale
        return {
            "panel_left": panel_left,
            "panel_top": panel_top,
            "base_x": base_x,
            "base_y": base_y,
            "label_x": label_x,
            "slider_x": slider_x,
            "scale": panel_scale,
        }

    def _update_slider(
        self,
        slider_id: str,
        slider: SliderState,
        x: float,
        y: float,
        rect_on: rl.Texture2D,
        rect_off: rl.Texture2D,
        scale: float,
    ) -> bool:
        rect_w = float(rect_on.width) * scale
        rect_h = float(rect_on.height) * scale
        if rect_w <= 0.0 or rect_h <= 0.0:
            return False
        bar_w = rect_w * float(slider.max_value)
        bar_h = rect_h
        mouse = rl.get_mouse_position()
        hovered = x <= mouse.x <= x + bar_w and y <= mouse.y <= y + bar_h

        changed = False
        if hovered:
            if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT):
                slider.value = max(slider.min_value, slider.value - 1)
                changed = True
            if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT):
                slider.value = min(slider.max_value, slider.value + 1)
                changed = True

        mouse_down = rl.is_mouse_button_down(rl.MOUSE_BUTTON_LEFT)
        if hovered and rl.is_mouse_button_pressed(rl.MOUSE_BUTTON_LEFT):
            self._active_slider = slider_id
        if self._active_slider == slider_id and mouse_down:
            relative = float(mouse.x) - x
            idx = int(relative // rect_w) + 1
            if idx < slider.min_value:
                idx = slider.min_value
            if idx > slider.max_value:
                idx = slider.max_value
            if slider.value != idx:
                slider.value = idx
                changed = True
        if self._active_slider == slider_id and not mouse_down:
            self._active_slider = None

        return changed

    def _update_checkbox(self, x: float, y: float, scale: float) -> bool:
        check_on = self._check_on
        check_off = self._check_off
        if check_on is None or check_off is None:
            return False
        font = self._ensure_small_font()
        text_scale = 1.0 * scale
        label = "UI Info texts"
        label_w = measure_small_text_width(font, label, text_scale)
        rect_w = float(check_on.width) * scale + 6.0 * scale + label_w
        rect_h = max(float(check_on.height) * scale, font.cell_size * text_scale)
        mouse = rl.get_mouse_position()
        hovered = x <= mouse.x <= x + rect_w and y <= mouse.y <= y + rect_h
        if hovered and rl.is_mouse_button_pressed(rl.MOUSE_BUTTON_LEFT):
            self._ui_info_texts = not self._ui_info_texts
            return True
        return False

    def _update_controls_button(self, x: float, y: float, scale: float) -> bool:
        tex = self._button_tex
        if tex is None:
            return False
        w = float(tex.width) * scale
        h = float(tex.height) * scale
        mouse = rl.get_mouse_position()
        hovered = x <= mouse.x <= x + w and y <= mouse.y <= y + h
        if hovered and rl.is_mouse_button_pressed(rl.MOUSE_BUTTON_LEFT):
            return True
        return False

    def _draw_options_contents(self) -> None:
        assets = self._assets
        if assets is None:
            return
        labels_tex = assets.labels
        layout = self._content_layout()
        base_x = layout["base_x"]
        base_y = layout["base_y"]
        label_x = layout["label_x"]
        slider_x = layout["slider_x"]
        scale = layout["scale"]

        font = self._ensure_small_font()
        text_scale = 1.0 * scale
        text_color = rl.Color(255, 255, 255, int(255 * 0.8))

        if labels_tex is not None:
            src = rl.Rectangle(
                0.0,
                float(MENU_LABEL_ROW_OPTIONS) * MENU_LABEL_ROW_HEIGHT,
                MENU_LABEL_WIDTH,
                MENU_LABEL_ROW_HEIGHT,
            )
            dst = rl.Rectangle(
                base_x,
                base_y,
                MENU_LABEL_WIDTH * scale,
                MENU_LABEL_ROW_HEIGHT * scale,
            )
            MenuView._draw_ui_quad(
                texture=labels_tex,
                src=src,
                dst=dst,
                origin=rl.Vector2(0.0, 0.0),
                rotation_deg=0.0,
                tint=rl.WHITE,
            )
        else:
            rl.draw_text(self._title, int(base_x), int(base_y), int(24 * scale), rl.WHITE)

        y_offsets = (47.0, 67.0, 87.0, 107.0)
        for label, offset in zip(self._LABELS, y_offsets, strict=False):
            draw_small_text(font, label, label_x, base_y + offset * scale, text_scale, text_color)

        rect_on = self._rect_on
        rect_off = self._rect_off
        if rect_on is None or rect_off is None:
            return
        rect_w = float(rect_on.width) * scale
        rect_h = float(rect_on.height) * scale

        self._draw_slider(self._slider_sfx, slider_x, base_y + 47.0 * scale, rect_on, rect_off, rect_w, rect_h)
        self._draw_slider(self._slider_music, slider_x, base_y + 67.0 * scale, rect_on, rect_off, rect_w, rect_h)
        self._draw_slider(self._slider_detail, slider_x, base_y + 87.0 * scale, rect_on, rect_off, rect_w, rect_h)
        self._draw_slider(self._slider_mouse, slider_x, base_y + 107.0 * scale, rect_on, rect_off, rect_w, rect_h)

        check_on = self._check_on
        check_off = self._check_off
        if check_on is not None and check_off is not None:
            check_tex = check_on if self._ui_info_texts else check_off
            check_w = float(check_tex.width) * scale
            check_h = float(check_tex.height) * scale
            check_x = label_x
            check_y = base_y + 135.0 * scale
            rl.draw_texture_pro(
                check_tex,
                rl.Rectangle(0.0, 0.0, float(check_tex.width), float(check_tex.height)),
                rl.Rectangle(check_x, check_y, check_w, check_h),
                rl.Vector2(0.0, 0.0),
                0.0,
                rl.WHITE,
            )
            draw_small_text(
                font,
                "UI Info texts",
                check_x + check_w + 6.0 * scale,
                check_y + 1.0 * scale,
                text_scale,
                text_color,
            )

        button = self._button_tex
        if button is not None:
            button_x = label_x - 8.0 * scale
            button_y = base_y + 155.0 * scale
            button_w = float(button.width) * scale
            button_h = float(button.height) * scale
            mouse = rl.get_mouse_position()
            hovered = button_x <= mouse.x <= button_x + button_w and button_y <= mouse.y <= button_y + button_h
            alpha = 255 if hovered else 220
            rl.draw_texture_pro(
                button,
                rl.Rectangle(0.0, 0.0, float(button.width), float(button.height)),
                rl.Rectangle(button_x, button_y, button_w, button_h),
                rl.Vector2(0.0, 0.0),
                0.0,
                rl.Color(255, 255, 255, alpha),
            )
            label = "Controls"
            label_w = measure_small_text_width(font, label, text_scale)
            text_x = button_x + (button_w - label_w) * 0.5
            text_y = button_y + (button_h - font.cell_size * text_scale) * 0.5
            draw_small_text(font, label, text_x, text_y, text_scale, rl.Color(20, 20, 20, 255))

    def _draw_slider(
        self,
        slider: SliderState,
        x: float,
        y: float,
        rect_on: rl.Texture2D,
        rect_off: rl.Texture2D,
        rect_w: float,
        rect_h: float,
    ) -> None:
        for idx in range(slider.max_value):
            tex = rect_on if idx < slider.value else rect_off
            dst = rl.Rectangle(x + float(idx) * rect_w, y, rect_w, rect_h)
            tint = rl.WHITE if idx < slider.value else rl.Color(255, 255, 255, int(255 * 0.5))
            rl.draw_texture_pro(
                tex,
                rl.Rectangle(0.0, 0.0, float(tex.width), float(tex.height)),
                dst,
                rl.Vector2(0.0, 0.0),
                0.0,
                tint,
            )



from __future__ import annotations

from pathlib import Path
import random
from typing import TYPE_CHECKING, Protocol

import pyray as rl

from grim.assets import PaqTextureCache
from grim.audio import AudioState, update_audio
from grim.console import ConsoleState
from grim.config import CrimsonConfig
from grim.fonts.small import SmallFontData, draw_small_text, load_small_font, measure_small_text_width
from grim.view import ViewContext

from ..gameplay import _creature_find_in_radius, perk_count_get
from ..game_world import GameWorld
from ..persistence.highscores import HighScoreRecord
from ..perks import PerkId
from ..ui.game_over import GameOverUi
from ..ui.hud import HudAssets, draw_target_health_bar, load_hud_assets

if TYPE_CHECKING:
    from ..persistence.save_status import GameStatus


def _clamp(value: float, lo: float, hi: float) -> float:
    if value < lo:
        return lo
    if value > hi:
        return hi
    return value


class _ScreenFade(Protocol):
    screen_fade_alpha: float


class BaseGameplayMode:
    def __init__(
        self,
        ctx: ViewContext,
        *,
        world_size: float,
        default_game_mode_id: int,
        demo_mode_active: bool = False,
        difficulty_level: int = 0,
        hardcore: bool = False,
        texture_cache: PaqTextureCache | None = None,
        config: CrimsonConfig | None = None,
        console: ConsoleState | None = None,
        audio: AudioState | None = None,
        audio_rng: random.Random | None = None,
    ) -> None:
        self._assets_root = ctx.assets_dir
        self._missing_assets: list[str] = []
        self._hud_missing: list[str] = []
        self._small: SmallFontData | None = None
        self._hud_assets: HudAssets | None = None

        self._default_game_mode_id = int(default_game_mode_id)
        self._config = config
        self._console = console
        self._base_dir = config.path.parent if config is not None else Path.cwd()

        self.close_requested = False
        self._action: str | None = None
        self._paused = False
        self._status: GameStatus | None = None

        self._world = GameWorld(
            assets_dir=ctx.assets_dir,
            world_size=float(world_size),
            demo_mode_active=bool(demo_mode_active),
            difficulty_level=int(difficulty_level),
            hardcore=bool(hardcore),
            texture_cache=texture_cache,
            config=config,
            audio=audio,
            audio_rng=audio_rng,
        )
        self._bind_world()

        self._game_over_active = False
        self._game_over_record: HighScoreRecord | None = None
        self._game_over_ui = GameOverUi(
            assets_root=self._assets_root,
            base_dir=self._base_dir,
            config=config or CrimsonConfig(path=self._base_dir / "crimson.cfg", data={"game_mode": int(default_game_mode_id)}),
        )
        self._game_over_banner = "reaper"

        self._ui_mouse_x = 0.0
        self._ui_mouse_y = 0.0
        self._cursor_pulse_time = 0.0
        self._last_dt_ms = 0.0
        self._screen_fade: _ScreenFade | None = None

    def _cvar_float(self, name: str, default: float = 0.0) -> float:
        console = self._console
        if console is None:
            return float(default)
        cvar = console.cvars.get(name)
        if cvar is None:
            return float(default)
        return float(cvar.value_f)

    def _hud_small_indicators(self) -> bool:
        return self._cvar_float("cv_uiSmallIndicators", 0.0) != 0.0

    def _config_game_mode_id(self) -> int:
        config = self._config
        if config is None:
            return int(self._default_game_mode_id)
        try:
            value = config.data.get("game_mode", self._default_game_mode_id)
            return int(value or self._default_game_mode_id)
        except Exception:
            return int(self._default_game_mode_id)

    def _draw_target_health_bar(self, *, alpha: float = 1.0) -> None:
        creatures = getattr(self._creatures, "entries", [])
        if not creatures:
            return

        if perk_count_get(self._player, PerkId.DOCTOR) <= 0:
            return

        target_idx = _creature_find_in_radius(
            creatures,
            pos_x=float(getattr(self._player, "aim_x", 0.0)),
            pos_y=float(getattr(self._player, "aim_y", 0.0)),
            radius=12.0,
            start_index=0,
        )
        if target_idx == -1:
            return

        creature = creatures[target_idx]
        if not bool(getattr(creature, "active", False)):
            return
        hp = float(getattr(creature, "hp", 0.0))
        max_hp = float(getattr(creature, "max_hp", 0.0))
        if hp <= 0.0 or max_hp <= 0.0:
            return

        ratio = hp / max_hp
        if ratio < 0.0:
            ratio = 0.0
        if ratio > 1.0:
            ratio = 1.0

        x0, y0 = self._world.world_to_screen(float(creature.x) - 32.0, float(creature.y) + 32.0)
        x1, _y1 = self._world.world_to_screen(float(creature.x) + 32.0, float(creature.y) + 32.0)
        width = float(x1) - float(x0)
        if width <= 1e-3:
            return
        draw_target_health_bar(
            x=float(x0),
            y=float(y0),
            width=width,
            ratio=ratio,
            alpha=float(alpha),
            scale=width / 64.0,
        )

    def _bind_world(self) -> None:
        self._state = self._world.state
        self._creatures = self._world.creatures
        self._player = self._world.players[0]
        self._state.status = self._status

    def bind_status(self, status: GameStatus | None) -> None:
        self._status = status
        self._state.status = status

    def bind_screen_fade(self, fade: _ScreenFade | None) -> None:
        self._screen_fade = fade

    def bind_audio(self, audio: AudioState | None, audio_rng: random.Random | None) -> None:
        self._world.audio = audio
        self._world.audio_rng = audio_rng

    def _update_audio(self, dt: float) -> None:
        if self._world.audio is not None:
            update_audio(self._world.audio, dt)

    def _ui_line_height(self, scale: float = 1.0) -> int:
        if self._small is not None:
            return int(self._small.cell_size * scale)
        return int(20 * scale)

    def _ui_text_width(self, text: str, scale: float = 1.0) -> int:
        if self._small is not None:
            return int(measure_small_text_width(self._small, text, scale))
        return int(rl.measure_text(text, int(20 * scale)))

    def _draw_ui_text(self, text: str, x: float, y: float, color: rl.Color, scale: float = 1.0) -> None:
        if self._small is not None:
            draw_small_text(self._small, text, x, y, scale, color)
        else:
            rl.draw_text(text, int(x), int(y), int(20 * scale), color)

    def _ui_mouse_pos(self) -> rl.Vector2:
        return rl.Vector2(float(self._ui_mouse_x), float(self._ui_mouse_y))

    def _update_ui_mouse(self) -> None:
        mouse = rl.get_mouse_position()
        screen_w = float(rl.get_screen_width())
        screen_h = float(rl.get_screen_height())
        self._ui_mouse_x = _clamp(float(mouse.x), 0.0, max(0.0, screen_w - 1.0))
        self._ui_mouse_y = _clamp(float(mouse.y), 0.0, max(0.0, screen_h - 1.0))

    def _tick_frame(self, dt: float, *, clamp_cursor_pulse: bool = False) -> tuple[float, float]:
        dt_frame = float(dt)
        dt_ui_ms = float(min(dt_frame, 0.1) * 1000.0)
        self._last_dt_ms = dt_ui_ms

        self._update_ui_mouse()

        pulse_dt = float(min(dt_frame, 0.1)) if clamp_cursor_pulse else dt_frame
        self._cursor_pulse_time += pulse_dt * 1.1

        return dt_frame, dt_ui_ms

    def _player_name_default(self) -> str:
        config = self._config
        if config is None:
            return ""
        raw = config.data.get("player_name")
        if isinstance(raw, (bytes, bytearray)):
            return bytes(raw).split(b"\x00", 1)[0].decode("latin-1", errors="ignore")
        if isinstance(raw, str):
            return raw
        return ""

    def open(self) -> None:
        self.close_requested = False
        self._action = None
        self._paused = False
        self._missing_assets.clear()
        self._hud_missing.clear()
        try:
            self._small = load_small_font(self._assets_root, self._missing_assets)
        except Exception:
            self._small = None

        self._hud_assets = load_hud_assets(self._assets_root)
        if self._hud_assets.missing:
            self._hud_missing = list(self._hud_assets.missing)

        self._game_over_active = False
        self._game_over_record = None
        self._game_over_banner = "reaper"
        self._game_over_ui.close()

        player_count = 1
        config = self._config
        if config is not None:
            try:
                player_count = int(config.data.get("player_count", 1) or 1)
            except Exception:
                player_count = 1
        seed = random.getrandbits(32)
        self._world.reset(seed=seed, player_count=max(1, min(4, player_count)))
        self._world.open()
        self._bind_world()

        self._ui_mouse_x = float(rl.get_screen_width()) * 0.5
        self._ui_mouse_y = float(rl.get_screen_height()) * 0.5
        self._cursor_pulse_time = 0.0

    def close(self) -> None:
        self._game_over_ui.close()
        if self._small is not None:
            rl.unload_texture(self._small.texture)
            self._small = None
        self._hud_assets = None
        self._world.close()

    def take_action(self) -> str | None:
        action = self._action
        self._action = None
        return action

    def _draw_screen_fade(self) -> None:
        fade_alpha = 0.0
        if self._screen_fade is not None:
            fade_alpha = float(self._screen_fade.screen_fade_alpha)
        if fade_alpha <= 0.0:
            return
        alpha = int(255 * max(0.0, min(1.0, fade_alpha)))
        rl.draw_rectangle(0, 0, int(rl.get_screen_width()), int(rl.get_screen_height()), rl.Color(0, 0, 0, alpha))



from __future__ import annotations

from dataclasses import dataclass, field

import pyray as rl


@dataclass(slots=True)
class ActionMap:
    bindings: dict[str, tuple[int, ...]] = field(default_factory=dict)

    def bind(self, action: str, *keys: int) -> None:
        if not keys:
            raise ValueError("bind requires at least one key")
        self.bindings[action] = tuple(int(key) for key in keys)

    def is_down(self, action: str) -> bool:
        keys = self.bindings.get(action, ())
        return any(rl.is_key_down(key) for key in keys)

    def was_pressed(self, action: str) -> bool:
        keys = self.bindings.get(action, ())
        return any(rl.is_key_pressed(key) for key in keys)


def is_key_down(key: int) -> bool:
    return rl.is_key_down(key)


def was_key_pressed(key: int) -> bool:
    return rl.is_key_pressed(key)


def is_mouse_button_down(button: int) -> bool:
    return rl.is_mouse_button_down(button)


def was_mouse_button_pressed(button: int) -> bool:
    return rl.is_mouse_button_pressed(button)


def mouse_position() -> tuple[int, int]:
    pos = rl.get_mouse_position()
    return int(pos.x), int(pos.y)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import random

from grim.audio import AudioState, init_audio_state
from grim.config import CrimsonConfig, ensure_crimson_cfg
from grim.console import ConsoleLog, ConsoleState

from ..assets_fetch import download_missing_paqs
from ..paths import default_runtime_dir


@dataclass(slots=True)
class ViewAudioBootstrap:
    config: CrimsonConfig | None
    console: ConsoleState | None
    audio: AudioState | None
    audio_rng: random.Random | None


def init_view_audio(assets_dir: Path, *, seed: int = 0xBEEF) -> ViewAudioBootstrap:
    runtime_dir = default_runtime_dir()
    runtime_dir.mkdir(parents=True, exist_ok=True)
    try:
        config = ensure_crimson_cfg(runtime_dir)
    except Exception:
        return ViewAudioBootstrap(None, None, None, None)

    console = ConsoleState(
        base_dir=runtime_dir,
        log=ConsoleLog(base_dir=runtime_dir),
        assets_dir=assets_dir,
    )
    try:
        download_missing_paqs(assets_dir, console)
    except Exception as exc:
        console.log.log(f"assets: download failed: {exc}")
        console.log.flush()

    try:
        audio = init_audio_state(config, assets_dir, console)
    except Exception:
        return ViewAudioBootstrap(config, console, None, None)

    return ViewAudioBootstrap(config, console, audio, random.Random(seed))



from __future__ import annotations

from dataclasses import dataclass
import random

import pyray as rl

from grim.assets import PaqTextureCache
from grim.audio import AudioState
from grim.console import ConsoleState
from grim.config import CrimsonConfig
from grim.view import ViewContext

from ..creatures.spawn import tick_rush_mode_spawns
from ..game_modes import GameMode
from ..gameplay import PlayerInput, most_used_weapon_id_for_player, weapon_assign_player
from ..input_codes import config_keybinds, input_code_is_down, input_code_is_pressed, player_move_fire_binds
from ..persistence.highscores import HighScoreRecord
from ..ui.cursor import draw_aim_cursor, draw_menu_cursor
from ..ui.hud import draw_hud_overlay, hud_flags_for_game_mode
from ..ui.perk_menu import load_perk_menu_assets
from .base_gameplay_mode import BaseGameplayMode

WORLD_SIZE = 1024.0
RUSH_WEAPON_ID = 2

UI_TEXT_SCALE = 1.0
UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_ERROR_COLOR = rl.Color(240, 80, 80, 255)


@dataclass(slots=True)
class _RushState:
    elapsed_ms: float = 0.0
    spawn_cooldown_ms: float = 0.0


class RushMode(BaseGameplayMode):
    def __init__(
        self,
        ctx: ViewContext,
        *,
        texture_cache: PaqTextureCache | None = None,
        config: CrimsonConfig | None = None,
        console: ConsoleState | None = None,
        audio: AudioState | None = None,
        audio_rng: random.Random | None = None,
    ) -> None:
        super().__init__(
            ctx,
            world_size=WORLD_SIZE,
            default_game_mode_id=int(GameMode.RUSH),
            demo_mode_active=False,
            difficulty_level=0,
            hardcore=False,
            texture_cache=texture_cache,
            config=config,
            console=console,
            audio=audio,
            audio_rng=audio_rng,
        )
        self._rush = _RushState()

        self._ui_assets = None

    def _enforce_rush_loadout(self) -> None:
        for player in self._world.players:
            if int(player.weapon_id) != RUSH_WEAPON_ID:
                weapon_assign_player(player, RUSH_WEAPON_ID)
            # `rush_mode_update` forces weapon+ammo every frame; keep ammo topped up.
            player.ammo = float(max(0, int(player.clip_size)))

    def open(self) -> None:
        super().open()
        self._ui_assets = load_perk_menu_assets(self._assets_root)
        if self._ui_assets.missing:
            self._missing_assets.extend(self._ui_assets.missing)
        self._rush = _RushState()
        self._enforce_rush_loadout()

    def close(self) -> None:
        if self._ui_assets is not None:
            self._ui_assets = None
        super().close()

    def _handle_input(self) -> None:
        if self._game_over_active:
            if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
                self._action = "back_to_menu"
                self.close_requested = True
            return

        if rl.is_key_pressed(rl.KeyboardKey.KEY_TAB):
            self._paused = not self._paused

        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self.close_requested = True

    def _build_input(self) -> PlayerInput:
        keybinds = config_keybinds(self._config)
        if not keybinds:
            keybinds = (0x11, 0x1F, 0x1E, 0x20, 0x100)
        up_key, down_key, left_key, right_key, fire_key = player_move_fire_binds(keybinds, 0)

        move_x = 0.0
        move_y = 0.0
        if input_code_is_down(left_key):
            move_x -= 1.0
        if input_code_is_down(right_key):
            move_x += 1.0
        if input_code_is_down(up_key):
            move_y -= 1.0
        if input_code_is_down(down_key):
            move_y += 1.0

        mouse = self._ui_mouse_pos()
        aim_x, aim_y = self._world.screen_to_world(float(mouse.x), float(mouse.y))

        fire_down = input_code_is_down(fire_key)
        fire_pressed = input_code_is_pressed(fire_key)

        return PlayerInput(
            move_x=move_x,
            move_y=move_y,
            aim_x=float(aim_x),
            aim_y=float(aim_y),
            fire_down=bool(fire_down),
            fire_pressed=bool(fire_pressed),
            reload_pressed=False,
        )

    def _player_name_default(self) -> str:
        config = self._config
        if config is None:
            return ""
        raw = config.data.get("player_name")
        if isinstance(raw, (bytes, bytearray)):
            return bytes(raw).split(b"\x00", 1)[0].decode("latin-1", errors="ignore")
        if isinstance(raw, str):
            return raw
        return ""

    def _enter_game_over(self) -> None:
        if self._game_over_active:
            return

        record = HighScoreRecord.blank()
        record.score_xp = int(self._player.experience)
        record.survival_elapsed_ms = int(self._rush.elapsed_ms)
        record.creature_kill_count = int(self._creatures.kill_count)
        weapon_id = most_used_weapon_id_for_player(self._state, player_index=int(self._player.index), fallback_weapon_id=int(self._player.weapon_id))
        record.most_used_weapon_id = int(weapon_id)
        fired = 0
        hit = 0
        try:
            fired = int(self._state.shots_fired[int(self._player.index)])
            hit = int(self._state.shots_hit[int(self._player.index)])
        except Exception:
            fired = 0
            hit = 0
        fired = max(0, int(fired))
        hit = max(0, min(int(hit), fired))
        record.shots_fired = fired
        record.shots_hit = hit
        record.game_mode_id = int(self._config.data.get("game_mode", int(GameMode.RUSH))) if self._config is not None else int(GameMode.RUSH)

        self._game_over_record = record
        self._game_over_ui.open()
        self._game_over_active = True

    def update(self, dt: float) -> None:
        self._update_audio(dt)

        dt_frame = self._tick_frame(dt)[0]
        self._handle_input()

        if self._game_over_active:
            record = self._game_over_record
            if record is None:
                self._enter_game_over()
                record = self._game_over_record
            if record is not None:
                action = self._game_over_ui.update(
                    dt,
                    record=record,
                    player_name_default=self._player_name_default(),
                    play_sfx=self._world.audio_router.play_sfx,
                    rand=self._state.rng.rand,
                    mouse=self._ui_mouse_pos(),
                )
                if action == "play_again":
                    self.open()
                    return
                if action == "high_scores":
                    self._action = "open_high_scores"
                    return
                if action == "main_menu":
                    self._action = "back_to_menu"
                    self.close_requested = True
            return

        any_alive = any(player.health > 0.0 for player in self._world.players)
        dt_world = 0.0 if self._paused or (not any_alive) else dt_frame

        self._rush.elapsed_ms += dt_world * 1000.0
        if dt_world <= 0.0:
            if not any_alive:
                self._enter_game_over()
            return

        self._enforce_rush_loadout()
        input_state = self._build_input()
        self._world.update(
            dt_world,
            inputs=[input_state for _ in self._world.players],
            auto_pick_perks=False,
            game_mode=int(GameMode.RUSH),
            perk_progression_enabled=False,
        )

        cooldown, spawns = tick_rush_mode_spawns(
            self._rush.spawn_cooldown_ms,
            dt_world * 1000.0,
            self._state.rng,
            player_count=len(self._world.players),
            survival_elapsed_ms=int(self._rush.elapsed_ms),
            terrain_width=float(self._world.world_size),
            terrain_height=float(self._world.world_size),
        )
        self._rush.spawn_cooldown_ms = cooldown
        self._creatures.spawn_inits(spawns)

        if not any(player.health > 0.0 for player in self._world.players):
            self._enter_game_over()

    def _draw_game_cursor(self) -> None:
        mouse_x = float(self._ui_mouse_x)
        mouse_y = float(self._ui_mouse_y)
        cursor_tex = self._ui_assets.cursor if self._ui_assets is not None else None
        draw_menu_cursor(
            self._world.particles_texture,
            cursor_tex,
            x=mouse_x,
            y=mouse_y,
            pulse_time=float(self._cursor_pulse_time),
        )

    def _draw_aim_cursor(self) -> None:
        mouse_x = float(self._ui_mouse_x)
        mouse_y = float(self._ui_mouse_y)
        aim_tex = self._ui_assets.aim if self._ui_assets is not None else None
        draw_aim_cursor(self._world.particles_texture, aim_tex, x=mouse_x, y=mouse_y)

    def draw(self) -> None:
        self._world.draw(draw_aim_indicators=(not self._game_over_active))
        self._draw_screen_fade()

        hud_bottom = 0.0
        if (not self._game_over_active) and self._hud_assets is not None:
            hud_flags = hud_flags_for_game_mode(self._config_game_mode_id())
            self._draw_target_health_bar()
            hud_bottom = draw_hud_overlay(
                self._hud_assets,
                player=self._player,
                players=self._world.players,
                bonus_hud=self._state.bonus_hud,
                elapsed_ms=self._rush.elapsed_ms,
                font=self._small,
                frame_dt_ms=self._last_dt_ms,
                show_health=hud_flags.show_health,
                show_weapon=hud_flags.show_weapon,
                show_xp=hud_flags.show_xp,
                show_time=hud_flags.show_time,
                show_quest_hud=hud_flags.show_quest_hud,
                small_indicators=self._hud_small_indicators(),
            )

        if not self._game_over_active:
            x = 18.0
            y = max(18.0, hud_bottom + 10.0)
            line = float(self._ui_line_height())
            self._draw_ui_text(f"rush: t={self._rush.elapsed_ms/1000.0:6.1f}s", x, y, UI_TEXT_COLOR)
            self._draw_ui_text(f"kills={self._creatures.kill_count}", x, y + line, UI_HINT_COLOR)
            if self._paused:
                self._draw_ui_text("paused (TAB)", x, y + line * 2.0, UI_HINT_COLOR)
            if self._player.health <= 0.0:
                self._draw_ui_text("game over", x, y + line * 2.0, UI_ERROR_COLOR)

        warn_y = float(rl.get_screen_height()) - 28.0
        if self._world.missing_assets:
            warn = "Missing world assets: " + ", ".join(self._world.missing_assets)
            self._draw_ui_text(warn, 24.0, warn_y, UI_ERROR_COLOR, scale=0.8)
            warn_y -= float(self._ui_line_height(scale=0.8)) + 2.0
        if self._hud_missing:
            warn = "Missing HUD assets: " + ", ".join(self._hud_missing)
            self._draw_ui_text(warn, 24.0, warn_y, UI_ERROR_COLOR, scale=0.8)

        if not self._game_over_active:
            self._draw_aim_cursor()
        else:
            self._draw_game_cursor()
            if self._game_over_record is not None:
                self._game_over_ui.draw(
                    record=self._game_over_record,
                    banner_kind=self._game_over_banner,
                    hud_assets=self._hud_assets,
                    mouse=self._ui_mouse_pos(),
                )



from __future__ import annotations

from dataclasses import dataclass
import math

from ..bonuses import BonusId
from ..creatures.damage import creature_apply_damage
from ..creatures.runtime import CreaturePool
from ..creatures.runtime import CREATURE_HITBOX_ALIVE
from ..creatures.anim import creature_anim_advance_phase
from ..creatures.spawn import CreatureFlags, CreatureTypeId, SpawnEnv
from ..effects import FxQueue, FxQueueRotated
from ..gameplay import (
    GameplayState,
    PlayerInput,
    PlayerState,
    bonus_update,
    perk_active,
    perks_update_effects,
    player_update,
    survival_progression_update,
)
from ..perks import PerkId
from ..player_damage import player_take_damage
from .world_defs import CREATURE_ANIM

ProjectileHit = tuple[int, float, float, float, float, float, float]


@dataclass(slots=True)
class WorldEvents:
    hits: list[ProjectileHit]
    deaths: tuple[object, ...]
    pickups: list[object]
    sfx: list[str]


@dataclass(slots=True)
class WorldState:
    spawn_env: SpawnEnv
    state: GameplayState
    players: list[PlayerState]
    creatures: CreaturePool

    @classmethod
    def build(
        cls,
        *,
        world_size: float,
        demo_mode_active: bool,
        hardcore: bool,
        difficulty_level: int,
    ) -> WorldState:
        spawn_env = SpawnEnv(
            terrain_width=float(world_size),
            terrain_height=float(world_size),
            demo_mode_active=bool(demo_mode_active),
            hardcore=bool(hardcore),
            difficulty_level=int(difficulty_level),
        )
        state = GameplayState()
        state.demo_mode_active = bool(demo_mode_active)
        state.hardcore = bool(hardcore)
        players: list[PlayerState] = []
        creatures = CreaturePool(env=spawn_env)
        return cls(
            spawn_env=spawn_env,
            state=state,
            players=players,
            creatures=creatures,
        )

    def step(
        self,
        dt: float,
        *,
        inputs: list[PlayerInput] | None,
        world_size: float,
        damage_scale_by_type: dict[int, float],
        detail_preset: int,
        fx_queue: FxQueue,
        fx_queue_rotated: FxQueueRotated,
        auto_pick_perks: bool,
        game_mode: int,
        perk_progression_enabled: bool,
    ) -> WorldEvents:
        dt = float(dt)
        if dt > 0.0 and self.players and perk_active(self.players[0], PerkId.REFLEX_BOOSTED):
            dt *= 0.9

        if inputs is None:
            inputs = [PlayerInput() for _ in self.players]

        prev_positions = [(player.pos_x, player.pos_y) for player in self.players]
        prev_health = [float(player.health) for player in self.players]

        # Native runs `perks_update_effects` early in the frame loop and relies on the current aim position
        # (`player_state_table.aim_x/aim_y`). Our aim is otherwise updated inside `player_update`, so stage it here.
        for idx, player in enumerate(self.players):
            input_state = inputs[idx] if idx < len(inputs) else PlayerInput()
            player.aim_x = float(input_state.aim_x)
            player.aim_y = float(input_state.aim_y)

        perks_update_effects(self.state, self.players, dt, creatures=self.creatures.entries, fx_queue=fx_queue)

        # `effects_update` runs early in the native frame loop, before creature/projectile updates.
        self.state.effects.update(dt, fx_queue=fx_queue)

        def _apply_projectile_damage_to_player(player_index: int, damage: float) -> None:
            idx = int(player_index)
            if not (0 <= idx < len(self.players)):
                return
            player_take_damage(self.state, self.players[idx], float(damage), dt=dt, rand=self.state.rng.rand)

        creature_result = self.creatures.update(
            dt,
            state=self.state,
            players=self.players,
            detail_preset=detail_preset,
            world_width=float(world_size),
            world_height=float(world_size),
            fx_queue=fx_queue,
            fx_queue_rotated=fx_queue_rotated,
        )

        deaths = list(creature_result.deaths)

        def _apply_projectile_damage_to_creature(
            creature_index: int,
            damage: float,
            damage_type: int,
            impulse_x: float,
            impulse_y: float,
            owner_id: int,
        ) -> None:
            idx = int(creature_index)
            if not (0 <= idx < len(self.creatures.entries)):
                return
            creature = self.creatures.entries[idx]
            if not creature.active:
                return

            death_start_needed = creature.hp > 0.0 and creature.hitbox_size == CREATURE_HITBOX_ALIVE
            killed = creature_apply_damage(
                creature,
                damage_amount=float(damage),
                damage_type=int(damage_type),
                impulse_x=float(impulse_x),
                impulse_y=float(impulse_y),
                owner_id=int(owner_id),
                dt=float(dt),
                players=self.players,
                rand=self.state.rng.rand,
            )
            if killed and death_start_needed:
                deaths.append(
                    self.creatures.handle_death(
                        idx,
                        state=self.state,
                        players=self.players,
                        rand=self.state.rng.rand,
                        detail_preset=int(detail_preset),
                        world_width=float(world_size),
                        world_height=float(world_size),
                        fx_queue=fx_queue,
                    )
                )

        hits = self.state.projectiles.update(
            dt,
            self.creatures.entries,
            world_size=float(world_size),
            damage_scale_by_type=damage_scale_by_type,
            detail_preset=int(detail_preset),
            rng=self.state.rng.rand,
            runtime_state=self.state,
            players=self.players,
            apply_player_damage=_apply_projectile_damage_to_player,
            apply_creature_damage=_apply_projectile_damage_to_creature,
        )
        self.state.secondary_projectiles.update_pulse_gun(
            dt,
            self.creatures.entries,
            apply_creature_damage=_apply_projectile_damage_to_creature,
            runtime_state=self.state,
            fx_queue=fx_queue,
            detail_preset=int(detail_preset),
        )

        for idx, player in enumerate(self.players):
            if idx >= len(prev_health):
                continue
            if float(prev_health[idx]) < 0.0:
                continue
            if float(player.health) >= 0.0:
                continue
            if not perk_active(player, PerkId.FINAL_REVENGE):
                continue

            px = float(player.pos_x)
            py = float(player.pos_y)
            rand = self.state.rng.rand
            self.state.effects.spawn_explosion_burst(
                pos_x=px,
                pos_y=py,
                scale=1.8,
                rand=rand,
                detail_preset=int(detail_preset),
            )

            prev_guard = bool(self.state.bonus_spawn_guard)
            self.state.bonus_spawn_guard = True
            for creature_idx, creature in enumerate(self.creatures.entries):
                if not creature.active:
                    continue
                if float(creature.hp) <= 0.0:
                    continue

                dx = float(creature.x) - px
                dy = float(creature.y) - py
                if abs(dx) > 512.0 or abs(dy) > 512.0:
                    continue

                remaining = 512.0 - math.hypot(dx, dy)
                if remaining <= 0.0:
                    continue

                damage = remaining * 5.0
                death_start_needed = float(creature.hp) > 0.0 and float(creature.hitbox_size) == CREATURE_HITBOX_ALIVE
                killed = creature_apply_damage(
                    creature,
                    damage_amount=damage,
                    damage_type=3,
                    impulse_x=0.0,
                    impulse_y=0.0,
                    owner_id=-1 - int(player.index),
                    dt=float(dt),
                    players=self.players,
                    rand=rand,
                )
                if killed and death_start_needed:
                    deaths.append(
                        self.creatures.handle_death(
                            int(creature_idx),
                            state=self.state,
                            players=self.players,
                            rand=rand,
                            detail_preset=int(detail_preset),
                            world_width=float(world_size),
                            world_height=float(world_size),
                            fx_queue=fx_queue,
                        )
                    )
            self.state.bonus_spawn_guard = prev_guard
            self.state.sfx_queue.append("sfx_explosion_large")
            self.state.sfx_queue.append("sfx_shockwave")

        def _kill_creature_no_corpse(creature_index: int, owner_id: int) -> None:
            idx = int(creature_index)
            if not (0 <= idx < len(self.creatures.entries)):
                return
            creature = self.creatures.entries[idx]
            if not creature.active:
                return
            if float(creature.hp) <= 0.0:
                return

            creature.last_hit_owner_id = int(owner_id)
            deaths.append(
                self.creatures.handle_death(
                    idx,
                    state=self.state,
                    players=self.players,
                    rand=self.state.rng.rand,
                    detail_preset=int(detail_preset),
                    world_width=float(world_size),
                    world_height=float(world_size),
                    fx_queue=fx_queue,
                    keep_corpse=False,
                )
            )

        self.state.particles.update(
            dt,
            creatures=self.creatures.entries,
            apply_creature_damage=_apply_projectile_damage_to_creature,
            kill_creature=_kill_creature_no_corpse,
        )
        self.state.sprite_effects.update(dt)

        for idx, player in enumerate(self.players):
            input_state = inputs[idx] if idx < len(inputs) else PlayerInput()
            player_update(player, input_state, dt, self.state, world_size=float(world_size))

        if dt > 0.0:
            self._advance_creature_anim(dt)
            self._advance_player_anim(dt, prev_positions)

        pickups = bonus_update(
            self.state,
            self.players,
            dt,
            creatures=self.creatures.entries,
            update_hud=True,
            apply_creature_damage=_apply_projectile_damage_to_creature,
            detail_preset=int(detail_preset),
        )
        if pickups:
            for pickup in pickups:
                if pickup.bonus_id != int(BonusId.NUKE):
                    self.state.effects.spawn_burst(
                        pos_x=float(pickup.pos_x),
                        pos_y=float(pickup.pos_y),
                        count=12,
                        rand=self.state.rng.rand,
                        detail_preset=detail_preset,
                        lifetime=0.4,
                        scale_step=0.1,
                        color_r=0.4,
                        color_g=0.5,
                        color_b=1.0,
                        color_a=0.5,
                    )
                if pickup.bonus_id == int(BonusId.REFLEX_BOOST):
                    self.state.effects.spawn_ring(
                        pos_x=float(pickup.pos_x),
                        pos_y=float(pickup.pos_y),
                        detail_preset=detail_preset,
                        color_r=0.6,
                        color_g=0.6,
                        color_b=1.0,
                        color_a=1.0,
                    )
                elif pickup.bonus_id == int(BonusId.FREEZE):
                    self.state.effects.spawn_ring(
                        pos_x=float(pickup.pos_x),
                        pos_y=float(pickup.pos_y),
                        detail_preset=detail_preset,
                        color_r=0.3,
                        color_g=0.5,
                        color_b=0.8,
                        color_a=1.0,
                    )

        if perk_progression_enabled:
            survival_progression_update(
                self.state,
                self.players,
                game_mode=game_mode,
                auto_pick=auto_pick_perks,
                dt=dt,
                creatures=self.creatures.entries,
            )

        sfx = list(creature_result.sfx)
        if self.state.sfx_queue:
            sfx.extend(self.state.sfx_queue)
            self.state.sfx_queue.clear()
        pain_sfx = ("sfx_trooper_inpain_01", "sfx_trooper_inpain_02", "sfx_trooper_inpain_03")
        death_sfx = ("sfx_trooper_die_01", "sfx_trooper_die_02")
        rand = self.state.rng.rand
        for idx, player in enumerate(self.players):
            if idx >= len(prev_health):
                continue
            before = float(prev_health[idx])
            after = float(player.health)
            if after >= before - 1e-6:
                continue
            if before <= 0.0:
                continue
            if after <= 0.0:
                # Prioritize death VO even if there are many other SFX this frame.
                sfx.insert(0, death_sfx[int(rand()) & 1])
            else:
                sfx.append(pain_sfx[int(rand()) % len(pain_sfx)])

        return WorldEvents(hits=hits, deaths=tuple(deaths), pickups=pickups, sfx=sfx)

    def _advance_creature_anim(self, dt: float) -> None:
        if float(self.state.bonuses.freeze) > 0.0:
            return
        for creature in self.creatures.entries:
            if not (creature.active and creature.hp > 0.0):
                continue
            try:
                type_id = CreatureTypeId(int(creature.type_id))
            except ValueError:
                continue
            info = CREATURE_ANIM.get(type_id)
            if info is None:
                continue
            creature.anim_phase, _ = creature_anim_advance_phase(
                creature.anim_phase,
                anim_rate=info.anim_rate,
                move_speed=float(creature.move_speed),
                dt=dt,
                size=float(creature.size),
                local_scale=float(getattr(creature, "move_scale", 1.0)),
                flags=creature.flags,
                ai_mode=int(creature.ai_mode),
            )

    def _advance_player_anim(self, dt: float, prev_positions: list[tuple[float, float]]) -> None:
        info = CREATURE_ANIM.get(CreatureTypeId.TROOPER)
        if info is None:
            return
        for idx, player in enumerate(self.players):
            if idx >= len(prev_positions):
                continue
            prev_x, prev_y = prev_positions[idx]
            speed = math.hypot(player.pos_x - prev_x, player.pos_y - prev_y)
            move_speed = speed / dt / 120.0 if dt > 0.0 else 0.0
            player.move_phase, _ = creature_anim_advance_phase(
                player.move_phase,
                anim_rate=info.anim_rate,
                move_speed=move_speed,
                dt=dt,
                size=float(player.size),
                local_scale=1.0,
                flags=CreatureFlags(0),
                ai_mode=0,
            )
